"""
// Copyright (C) 2022-2024 MLACS group (AC, CD, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import shutil
import pytest
from itertools import chain
from pathlib import Path


@pytest.fixture(autouse=True)
def root():
    return Path().absolute()


@pytest.fixture(autouse=True)
def expected_folder_base():
    folder = ["MolecularDynamics", "Snap"]
    return folder


@pytest.fixture(autouse=True)
def expected_files_base():
    files = ["MLACS.log", "Training_configurations.traj", "Trajectory.traj",
             "MLIP-Energy_comparison.dat", "MLIP-Forces_comparison.dat",
             "MLIP-Stress_comparison.dat", "Trajectory_potential.dat"]
    return files


@pytest.fixture(autouse=True)
def treelink(root, expected_folder, expected_files):

    for folder in expected_folder:
        if (root/folder).exists():
            shutil.rmtree(root / folder)

    for f in expected_files:
        if (root/f).exists():
            (root / f).unlink()

    folder, files = expected_folder, expected_files
    yield dict(folder=folder, files=files)

    for folder in expected_folder:
        shutil.rmtree(root / folder)

    for f in expected_files:
        (root / f).unlink()


@pytest.fixture
def langevin_treelink(root, expected_folder, expected_files):
    if "MolecularDynamics" in expected_folder:
        expected_folder.remove("MolecularDynamics")

    for folder in expected_folder:
        if (root/folder).exists():
            shutil.rmtree(root / folder)

    for f in expected_files:
        if (root/f).exists():
            (root / f).unlink()

    folder, files = expected_folder, expected_files
    yield dict(folder=folder, files=files)

    for folder in expected_folder:
        shutil.rmtree(root / folder)

    for f in expected_files:
        (root / f).unlink()


@pytest.fixture(autouse=True)
def clean_up_nc(root):
    def _clean_up():
        patterns = [root.rglob("*_WEIGHTS.nc"), root.rglob("*_HIST.nc")]
        for filename in chain(*patterns):
            if 'reference_files' in str(filename):
                continue
            filename.unlink()
    _clean_up()
    yield
    _clean_up()



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from unittest.mock import patch

import numpy as np
from ase.build import bulk
from ase.calculators.singlepoint import SinglePointCalculator

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.mlip.mlip_manager import MlipManager, SelfMlipManager
from mlacs.mlip.delta_learning import DeltaLearningPotential


def create_dum_data(atoms):
    at = atoms.copy()
    energy = np.random.random()
    forces = np.random.random((len(atoms), 3))
    stress = np.random.random(6)
    calc = SinglePointCalculator(atoms, energy=energy, forces=forces,
                                 stress=stress)
    at.calc = calc
    return at


@patch.multiple(MlipManager, __abstractmethods__=set())
def test_update_matrices():
    at = bulk("Si")
    desc = SnapDescriptor(at, 2.4)
    manager = MlipManager(desc)

    nconfs = 0
    nparams = desc.ncolumns
    assert manager.nconfs == nconfs

    # Let's try adding one atom
    fakeat = create_dum_data(at)
    manager.update_matrices(fakeat)
    nconfs += 1
    assert manager.nconfs == nconfs

    nf = 3 * len(at) * nconfs
    ns = nconfs * 6
    assert manager.amat_e.shape == (nconfs, nparams)
    assert manager.amat_f.shape == (nf, nparams)
    assert manager.amat_s.shape == (ns, nparams)
    assert manager.ymat_e.shape == (nconfs,)
    assert manager.ymat_f.shape == (nf,)
    assert manager.ymat_s.shape == (ns,)

    # Let's try adding several atoms
    fakeat = []
    for _ in range(5):
        fakeat.append(create_dum_data(at))
        nconfs += 1
    manager.update_matrices(fakeat)
    assert manager.nconfs == nconfs

    nf = 3 * len(at) * nconfs
    ns = nconfs * 6
    assert manager.amat_e.shape == (nconfs, nparams)
    assert manager.amat_f.shape == (nf, nparams)
    assert manager.amat_s.shape == (ns, nparams)
    assert manager.ymat_e.shape == (nconfs,)
    assert manager.ymat_f.shape == (nf,)
    assert manager.ymat_s.shape == (ns,)

    # if manager.folder.exists():
    #    shutil.rmtree(manager.folder)


@patch.multiple(SelfMlipManager, __abstractmethods__=set())
def test_update_matrices_self():
    at = bulk("Si")
    desc = SnapDescriptor(at, 2.3)
    manager = SelfMlipManager(desc)

    nconfs = 0
    assert len(manager.natoms) == nconfs

    # Let's try adding one atom
    fakeat = create_dum_data(at)
    manager.update_matrices(fakeat)
    nconfs += 1
    assert manager.nconfs == nconfs
    assert len(manager.natoms) == nconfs
    assert len(manager.configurations) == nconfs

    # Let's try adding several atoms
    fakeat = []
    for _ in range(5):
        fakeat.append(create_dum_data(at))
        nconfs += 1
    manager.update_matrices(fakeat)
    assert manager.nconfs == nconfs
    assert len(manager.natoms) == nconfs
    assert len(manager.configurations) == nconfs

    # Now check that the atoms are the same we added
    manager = SelfMlipManager(desc)
    manager.update_matrices(fakeat)
    for refat, predat in zip(fakeat, manager.configurations):
        assert refat == predat


@patch.multiple(SelfMlipManager, __abstractmethods__=set())
def test_update_matrices_delta():
    at = bulk("Si")
    desc = SnapDescriptor(at, 2.3)
    nparams = desc.ncolumns
    model = LinearPotential(desc, parameters=dict(twojmax=4))
    ref_pair_style = "zbl 3.0 4.0"
    ref_pair_coeff = ["* * 14 14"]
    manager = DeltaLearningPotential(model, ref_pair_style,
                                     ref_pair_coeff)

    nconfs = 0
    assert len(manager.model.natoms) == nconfs

    # Let's try adding one atom
    fakeat = create_dum_data(at)
    manager.update_matrices(fakeat)
    nconfs += 1
    assert manager.model.nconfs == nconfs
    assert len(manager.model.natoms) == nconfs

    nf = 3 * len(at) * nconfs
    ns = nconfs * 6
    assert manager.model.amat_e.shape == (nconfs, nparams)
    assert manager.model.amat_f.shape == (nf, nparams)
    assert manager.model.amat_s.shape == (ns, nparams)
    assert manager.model.ymat_e.shape == (nconfs,)
    assert manager.model.ymat_f.shape == (nf,)
    assert manager.model.ymat_s.shape == (ns,)

    # Let's try adding several atoms
    fakeat = []
    for _ in range(5):
        fakeat.append(create_dum_data(at))
        nconfs += 1
    manager.update_matrices(fakeat)
    assert manager.nconfs == nconfs
    assert len(manager.model.natoms) == nconfs

    nf = 3 * len(at) * nconfs
    ns = nconfs * 6
    assert manager.model.amat_e.shape == (nconfs, nparams)
    assert manager.model.amat_f.shape == (nf, nparams)
    assert manager.model.amat_s.shape == (ns, nparams)
    assert manager.model.ymat_e.shape == (nconfs,)
    assert manager.model.ymat_f.shape == (nf,)
    assert manager.model.ymat_s.shape == (ns,)

    # if manager.folder.exists():
    #    shutil.rmtree(manager.folder)



"""
// Copyright (C) 2022-2024 MLACS group (AC, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import shutil
import shlex
from pathlib import Path
from subprocess import run, PIPE

import numpy as np
from ase.atoms import Atoms
from ase.io.lammpsdata import write_lammps_data

from ..core.manager import Manager
from ..utilities import get_elements_Z_and_masses
from .mliap_descriptor import default_snap
from .descriptor import Descriptor, combine_reg
from ..utilities.io_lammps import (LammpsInput, LammpsBlockInput,
                                   get_lammps_command)


# ========================================================================== #
# ========================================================================== #
class SnapDescriptor(Descriptor):
    """
    Interface to the SNAP potential of LAMMPS.

    Parameters
    ----------
    atoms : :class:`ase.atoms`
        Reference structure, with the elements for the descriptor

    rcut: :class:`float`
        The cutoff of the descriptor, in angstrom
        Default 5.0

    parameters: :class:`dict`
        A dictionnary of parameters for the descriptor input

        The default values are

        - twojmax = 8
        - rfac0 = 0.99363
        - rmin0 = 0.0
        - switchflag = 1
        - bzeroflag = 1
        - wselfallflag = 0

    model: :class:`str`
        The type of model use. Can be either 'linear' or 'quadratic'
        Default `linear`

    alpha: :class:`float`
        The multiplication factor to the regularization parameter for
        ridge regression.
        Default 1.0

    alpha_quad: :class:`float`
        A multiplication factor for the regularization that apply only to
        the quadratic component of the descriptor
        Default 1.0

    Examples
    --------

    >>> from ase.build import bulk
    >>> atoms = bulk("Cu", cubic=True).repeat(2)
    >>>
    >>> from mlacs.mlip import SnapDescriptor, LinearPotential
    >>> desc = SnapDescriptor(atoms, rcut=4.2, parameters=dict(twojmax=6))
    >>> desc.compute_descriptor(atoms)
    """
    def __init__(self, atoms, rcut=5.0, parameters=dict(),
                 model="linear", alpha=1.0, alpha_quad=1.0, prefix='SNAP',
                 **kwargs):
        self.chemflag = parameters.pop("chemflag", 0)
        Descriptor.__init__(self, atoms, rcut, alpha, prefix=prefix, **kwargs)
        self.alpha_quad = alpha_quad
        self.model = model

        # Initialize the parameters for the descriptors
        self.radelems = parameters.pop("radelems", None)
        if self.radelems is None:
            self.radelems = np.array([0.5 for i in self.elements])
        self.welems = parameters.pop("welems", None)
        if self.welems is None:
            self.welems = np.array(self.Z) / np.sum(self.Z)
        self.params = default_snap
        self.params.update(parameters)
        if self.chemflag:
            self.params["bnormflag"] = 1
        twojmax = self.params["twojmax"]
        if twojmax % 2 == 0:
            m = 0.5 * twojmax + 1
            self.ndesc = int(m * (m+1) * (2*m+1) / 6)
        else:
            m = 0.5 * (twojmax + 1)
            self.ndesc = int(m * (m+1) * (m+2) / 3)
        if self.chemflag:
            self.ndesc *= self.nel**3
        if self.model == "quadratic":
            self.ndesc_quad = int(self.ndesc * (self.ndesc + 1) / 2)
            self.ndesc_lin = self.ndesc
            self.ndesc += int(self.ndesc * (self.ndesc + 1) / 2)
        self.ncolumns = int(self.nel * (self.ndesc + 1))

        self.cmd = get_lammps_command()

# ========================================================================== #
    def compute_descriptor(self, atoms, forces=True, stress=True):
        """
        """
        nat = len(atoms)
        el, _, _, _ = get_elements_Z_and_masses(atoms)
        chemsymb = np.array(atoms.get_chemical_symbols())

        lmp_atfname = "atoms.lmp"
        self._write_lammps_input(atoms.get_pbc())
        self._write_mlip_params()

        amat_e = np.zeros((1, self.ncolumns))
        amat_f = np.zeros((3 * nat, self.ncolumns))
        amat_s = np.zeros((6, self.ncolumns))

        write_lammps_data(lmp_atfname,
                          atoms,
                          specorder=self.elements.tolist())
        self._run_lammps(lmp_atfname)

        bispectrum = np.loadtxt("descriptor.out",
                                skiprows=4)
        bispectrum[-6:, 1:-1] /= -atoms.get_volume()

        amat_e[0, self.nel:] = bispectrum[0, 1:-1]
        amat_f[:, self.nel:] = bispectrum[1:3*nat+1, 1:-1]
        amat_s[:, self.nel:] = bispectrum[3*nat+1:, 1:-1]

        for i, el in enumerate(self.elements):
            amat_e[0, i] = np.count_nonzero(chemsymb == el)

        self.cleanup()
        res = dict(desc_e=amat_e,
                   desc_f=amat_f,
                   desc_s=amat_s)
        return res

# ========================================================================== #
    @Manager.exec_from_path
    def compute_descriptors(self, atoms, forces=True, stress=True):
        """
        Compute the descriptors of multiples structures in one Lammps launch
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if len(atoms) == 1:
            return super().compute_descriptors(atoms,
                                               forces=forces,
                                               stress=stress)
        if Path("Atoms").exists():
            shutil.rmtree("Atoms")
        Path("Atoms").mkdir()

        for i, at in enumerate(atoms):
            if np.any(at.get_pbc() != atoms[0].get_pbc()):
                raise ValueError("PBC cannot change between states")
            lmp_atfname = f"Atoms/atoms{i+1}.lmp"
            write_lammps_data(lmp_atfname,
                              at,
                              specorder=self.elements.tolist())
        self._write_mlip_params()
        self._write_looping_lammps_input(atoms)

        self._run_lammps(lmp_atfname)

        desc = []
        for idx in range(len(atoms)):
            nat = len(atoms[idx])
            chemsymb = np.array(atoms[idx].get_chemical_symbols())

            amat_e = np.zeros((1, self.ncolumns))
            amat_f = np.zeros((3 * nat, self.ncolumns))
            amat_s = np.zeros((6, self.ncolumns))

            bispectrum = np.loadtxt(f"descriptor{idx+1}.out",
                                    skiprows=4)
            bispectrum[-6:, 1:-1] /= -atoms[idx].get_volume()
            amat_e[0, self.nel:] = bispectrum[0, 1:-1]
            amat_f[:, self.nel:] = bispectrum[1:3*nat+1, 1:-1]
            amat_s[:, self.nel:] = bispectrum[3*nat+1:, 1:-1]

            for i, el in enumerate(self.elements):
                amat_e[0, i] = np.count_nonzero(chemsymb == el)

            desc.append(dict(desc_e=amat_e,
                             desc_f=amat_f,
                             desc_s=amat_s))
        self.loop_cleanup(len(atoms))
        return desc

# ========================================================================== #
    @Manager.exec_from_path
    def loop_cleanup(self, nconf):
        '''
        Function to cleanup the LAMMPS files used
        to extract the descriptor and gradient values
        '''
        Path("lmp.out").unlink(missing_ok=True)
        Path("lammps_input.in").unlink(missing_ok=True)
        shutil.rmtree("Atoms")
        for i in range(nconf):
            Path(f"descriptor{i+1}.out").unlink(missing_ok=True)

# ========================================================================== #
    @Manager.exec_from_path
    def _write_looping_lammps_input(self, atoms):
        """
        Write one lammps file for all the config in atoms
        """
        txt = "LAMMPS input file for extracting MLIP descriptors"
        lmp_in = LammpsInput(txt)

        block = LammpsBlockInput("init", "Initialization")
        block("loop", f"variable a loop {len(atoms)}")
        block("label", "label loopstart")

        pbc_txt = "{0} {1} {2}".format(
                *tuple("sp"[int(x)] for x in atoms[0].get_pbc()))
        block("boundary", f"boundary {pbc_txt}")
        block("atom_style", "atom_style  atomic")
        block("units", "units metal")
        block("read_data", "read_data Atoms/atoms${a}.lmp")
        for i, m in enumerate(self.masses):
            block(f"mass{i}", f"mass   {i+1} {m}")
        lmp_in("init", block)

        block = LammpsBlockInput("interaction", "Interactions")
        block("pair_style", f"pair_style zero {2*self.rcut}")
        block("pair_coeff", "pair_coeff  * *")
        lmp_in("interaction", block)

        block = LammpsBlockInput("fake_dynamic", "Fake dynamic")
        block("thermo", "thermo 100")
        block("timestep", "timestep 0.005")
        block("neighbor", "neighbor 1.0 bin")
        block("neigh_modify", "neigh_modify once no every 1 delay 0 check yes")
        lmp_in("fake_dynamic", block)

        block = LammpsBlockInput("compute", "Compute")
        block("compute", f"compute ml all snap {self._snap_opt_str()}")
        block("fix", "fix ml all ave/time 1 1 1 c_ml[*] " +
              "file descriptor${a}.out mode vector")
        block("run", "run 0")
        lmp_in("compute", block)

        block = LammpsBlockInput("loopback", "Loop Back")
        block("clear data", "clear")
        block("iterate", "next a")
        block("loopend", "jump SELF loopstart")
        lmp_in("loopback", block)

        with open("lammps_input.in", "w") as fd:
            fd.write(str(lmp_in))

# ========================================================================== #
    def _write_lammps_input(self, pbc):
        """
        """
        txt = "LAMMPS input file for extracting SNAP descriptors"
        lmp_in = LammpsInput(txt)

        block = LammpsBlockInput("init", "Initialization")
        block("clear", "clear")
        pbc_txt = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))
        block("boundary", f"boundary {pbc_txt}")
        block("atom_style", "atom_style  atomic")
        block("units", "units metal")
        block("read_data", "read_data atoms.lmp")
        for i, m in enumerate(self.masses):
            block(f"mass{i}", f"mass   {i+1} {m}")
        lmp_in("init", block)

        block = LammpsBlockInput("interaction", "Interactions")
        block("pair_style", f"pair_style zero {2*self.rcut}")
        block("pair_coeff", "pair_coeff  * *")
        lmp_in("interaction", block)

        block = LammpsBlockInput("fake_dynamic", "Fake dynamic")
        block("thermo", "thermo 100")
        block("timestep", "timestep 0.005")
        block("neighbor", "neighbor 1.0 bin")
        block("neigh_modify", "neigh_modify once no every 1 delay 0 check yes")
        lmp_in("fake_dynamic", block)

        block = LammpsBlockInput("compute", "Compute")
        block("compute", f"compute ml all snap {self._snap_opt_str()}")
        block("fix", "fix ml all ave/time 1 1 1 c_ml[*] " +
              "file descriptor.out mode vector")
        block("run", "run 0")
        lmp_in("compute", block)

        with open("lammps_input.in", "w") as fd:
            fd.write(str(lmp_in))

# ========================================================================== #
    def _run_lammps(self, lmp_atoms_fname):
        '''
        Function that call LAMMPS to extract the descriptor and gradient values
        '''
        lmp_cmd = f"{self.cmd} -in lammps_input.in -log none -sc lmp.out"
        lmp_handle = run(shlex.split(lmp_cmd),
                         stderr=PIPE)

        # There is a bug in LAMMPS that makes compute_mliap crashes at the end
        if lmp_handle.returncode != 0:
            msg = "LAMMPS stopped with the exit code \n" + \
                  f"{lmp_handle.stderr.decode()}"
            raise RuntimeError(msg)

# ========================================================================== #
    def cleanup(self):
        '''
        Function to cleanup the LAMMPS files used
        to extract the descriptor and gradient values
        '''
        Path("lmp.out").unlink()
        Path("descriptor.out").unlink()
        Path("lammps_input.in").unlink()
        Path("atoms.lmp").unlink()

# ========================================================================== #
    @Manager.exec_from_subdir
    def _write_mlip_params(self):
        """
        Function to write the mliap.descriptor parameter files of the MLIP
        """
        fname = f"{self.prefix}.descriptor"
        with open(fname, "w") as f:
            f.write(self.get_mlip_params())

# ========================================================================== #
    def get_mlip_params(self):
        s = ("# ")
        # Adding a commment line to know what elements are fitted here
        for elements in self.elements:
            s += ("{:} ".format(elements))
        s += ("MLIP parameters\n")
        s += ("# Descriptor:  SNAP\n")
        s += (f"# Model:       {self.model}\n")
        s += ("\n")
        s += (f"rcutfac         {self.rcut}\n")
        for key in self.params.keys():
            s += (f"{key:12}    {self.params[key]}\n")
        if self.chemflag:
            s += ("\n\n")
            s += ("chemflag     1\n")
            s += ("bnormflag    1\n")
        if self.model == "quadratic":
            s += ("quadraticflag  1")
        return s

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def write_mlip(self, coefficients):
        """
        """
        filepath = Path(self.get_filepath('.model'))
        if filepath.exists():
            filepath.unlink()
        fname = filepath.relative_to(self.subsubdir)

        intercepts = coefficients[:self.nel]
        coefs = coefficients[self.nel:]
        with open(filepath, "w") as fd:
            fd.write("# ")
            fd.write(" ".join(self.elements))
            fd.write(" MLIP parameters\n")
            fd.write("# Descriptor   SNAP\n")
            fd.write("\n")
            fd.write(f"{self.nel} {self.ndesc+1}\n")

            for iel in range(self.nel):
                iidx = iel * self.ndesc
                fidx = (iel + 1) * self.ndesc
                el = self.elements[iel]
                rel = self.radelems[iel]
                wel = self.welems[iel]
                fd.write(f"{el} {rel} {wel}\n")
                fd.write(f"{intercepts[iel]:35.30f}\n")
                np.savetxt(fd, coefs[iidx:fidx], fmt="%35.30f")

        return fname

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def get_coef(self, filename=None):
        """
        Read MLIP coefficients from a file.
        """
        if filename:
            filename = Path(filename)
        else:
            filename = Path(self.get_filepath('.model'))

        if not filename.is_file():
            filename = filename.absolute()
            raise FileNotFoundError(f"File {filename} does not exist")

        with open(filename, "r") as fd:
            lines = fd.readlines()

        coefs = []
        coefficients = []
        intercepts = []
        for line in lines:
            line = line.strip()
            if line.startswith('#') or len(line) == 0:
                continue
            line = line.split()
            if len(line) == 2:  # Consistency check: nel, ndesc+1
                assert int(line[0]) == self.nel, "The descriptor changed"
                assert int(line[1]) == self.ndesc+1, "The descriptor changed"
                continue

            if len(line) == 3:  # el, radelems, welems
                continue
            if (len(coefs)+len(intercepts)) % (self.ndesc+1) == 0:
                intercepts.append(float(line[0]))
            else:
                coefs.append(float(line[0]))
        coefficients = np.r_[intercepts, coefs]
        return coefficients

# ========================================================================== #
    def _regularization_matrix(self):
        # no regularization for the intercept
        d2 = [np.zeros((self.nel, self.nel))]
        if self.model == "linear":
            d2.append(np.eye(self.ndesc) * self.alpha)
        elif self.model == "quadratic":
            for i in range(self.nel):
                d2.append(np.eye(self.ndesc_lin) * self.alpha)
                d2.append(np.eye(self.ndesc_quad) * self.alpha_quad)
        return combine_reg(d2)

# ========================================================================== #
    def get_pair_style(self):
        return "snap"

# ========================================================================== #
    def get_pair_coeff(self):
        modelfile = self.get_filepath('.model')
        descfile = self.subdir / f"{self.prefix}.descriptor"
        pair_coeff = [f"* * {modelfile}  {descfile} " +
                      ' '.join(self.elements)]
        return pair_coeff

# ========================================================================== #
    def get_pair_style_coeff(self):
        return self.get_pair_style(), self.get_pair_coeff()

# ========================================================================== #
    def _snap_opt_str(self):
        params = self.params.copy()
        rfac0 = params.pop("rfac0")
        twojmax = params.pop("twojmax")
        radelems = [str(rad) for rad in self.radelems]
        welems = [str(wel) for wel in self.welems]
        quadflag = 0
        if self.model == "quadratic":
            quadflag = 1
        bnormflag = 0
        if self.chemflag:
            bnormflag = 1

        txt = f"{self.rcut} {rfac0} {twojmax} "
        txt += ' '.join(radelems) + " "
        txt += ' '.join(welems) + " "
        txt += f"quadraticflag {quadflag} "
        if self.chemflag:
            txt += f"chem {self.nel} "
            for i in range(self.nel):
                txt += f"{i} "
            txt += f"bnormflag {bnormflag} "
        for key, val in params.items():
            txt += f"{key} {val} "
        return txt

# ========================================================================== #
    def __str__(self):
        txt = " ".join(self.elements)
        txt += " SNAP descriptor,"
        txt += f" rcut = {self.rcut}"
        return txt

# ========================================================================== #
    def __repr__(self):
        txt = "SNAP descriptor\n"
        txt += f"{(len(txt) - 1) * '-'}\n"
        txt += "Elements :\n"
        txt += " ".join(self.elements) + "\n"
        txt += "Parameters :\n"
        txt += f"rcut                {self.rcut}\n"
        txt += f"chemflag            {self.chemflag}\n"
        for key, val in self.params.items():
            txt += f"{key:12}        {val}\n"
        txt += f"dimension           {self.ncolumns}\n"
        return txt



"""
// Copyright (C) 2022-2024 MLACS group (PR, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import numpy as np

from .thermostate import ThermoState
from ..core.manager import Manager
from ..utilities.thermo import (free_energy_uhlenbeck_ford,
                                free_energy_ideal_gas)

from ..utilities.io_lammps import LammpsBlockInput

p_tabled = [1, 25, 50, 75, 100]


# ========================================================================== #
# ========================================================================== #
class UFLiquidState(ThermoState):
    """
    Class for performing thermodynamic integration
    from a Uhlenbeck-Ford potential reference

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        ASE atoms object on which the simulation will be performed

    pair_style: :class:`str`
        pair_style for the LAMMPS input

    pair_coeff: :class:`str` or :class:`list` of :class:`str`
        pair_coeff for the LAMMPS input

    temperature: :class:`float`
        Temperature of the simulation

    pressure: :class:float
        Pressure. None default value

    fcorr1: :class:`float` or ``None``
        First order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    fcorr2: :class:`float` or ``None``
        Second order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    p: :class:`int`
        p parameter of the Uhlenbeck-Ford potential.
        Should be ``1``, ``25``, ``50``, ``75`` or ``100``. Default ``50``

    sigma: :class:`float`
        sigma parameter of the Uhlenbeck-Ford potential. Default ``2.0``.

    dt: :class:`int` (optional)
        Timestep for the simulations, in fs. Default ``1``

    damp : :class:`float` (optional)
        Damping parameter.
        If ``None``, a damping parameter of 100 x dt is used.

    pdamp: :class:`float` (optional)
        Pressure damping parameter, used is the pressure is not `None`
        By default, this correspond to 1000 times the timestep.

    nsteps: :class:`int` (optional)
        Number of production steps. Default ``10000``.

    nsteps_eq: :class:`int` (optional)
        Number of equilibration steps. Default ``5000``.

    nsteps_averaging: :class:`int` (optional)
        Number of step for equilibrate ideal structure
        at zero or finite pressure. Default ``10000``.

    rng: :class:`RNG object`
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    langevin: :class:`Bool`
        Settle or not a langevin thermostat to equilibrate
        an ideal structure at zero or finite pressure. Default ``True``

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    interval : :class:`int` (optional)
        Number of steps between log and traj writing. Override
        loginterval. Default ``50``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.
    """
    def __init__(self,
                 atoms,
                 pair_style,
                 pair_coeff,
                 temperature,
                 pressure=None,
                 fcorr1=None,
                 fcorr2=None,
                 p=50,
                 sigma=2.0,
                 dt=1,
                 damp=None,
                 pdamp=None,
                 nsteps=10000,
                 nsteps_eq=5000,
                 nsteps_averaging=10000,
                 rng=None,
                 langevin=True,
                 logfile=None,
                 trajfile=None,
                 interval=500,
                 loginterval=50,
                 **kwargs):

        kwargs.setdefault('folder', f"LiquidUF_T{temperature}K")

        super().__init__(atoms,
                         pair_style,
                         pair_coeff,
                         dt=dt,
                         nsteps=nsteps,
                         nsteps_eq=nsteps_eq,
                         nsteps_averaging=nsteps_averaging,
                         rng=rng,
                         langevin=langevin,
                         logfile=logfile,
                         trajfile=trajfile,
                         interval=interval,
                         loginterval=loginterval,
                         **kwargs)

        self.atoms = atoms
        self.temperature = temperature
        self.pressure = pressure
        self.damp = damp
        self.pdamp = pdamp

        self.fcorr1 = fcorr1
        self.fcorr2 = fcorr2

        self.p = p
        self.sigma = sigma
        self.dt = dt
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()
        self.langevin = langevin

        if self.pressure is not None:
            self.equilibrate = True
        else:
            self.equilibrate = False

        self.nsteps = nsteps
        self.nsteps_eq = nsteps_eq

        if self.p not in p_tabled:
            msg = "The p value of the UF potential has to be one for " + \
                  "which the free energy of the Uhlenbeck-Ford potential " + \
                  "is tabulated\n" + \
                  "Those value are : 1, 25, 50, 75 and 100"
            raise ValueError(msg)

# ========================================================================== #
    @Manager.exec_from_path
    def run(self):
        """
        """

        if self.equilibrate:
            self.run_averaging()

        self.run_dynamics(self.atoms, self.pair_style, self.pair_coeff)

        with open("MLMD.done", "w") as f:
            f.write("Done")

# ========================================================================== #
    @Manager.exec_from_path
    def postprocess(self):
        """
        Compute the free energy from the simulation
        """
        pass
        # Get needed value/constants
        vol = self.atoms.get_volume()  # angs**3
        nat_tot = len(self.atoms)

        nat = []
        for iel, e in enumerate(self.elem):
            nat.append(np.count_nonzero([a == e for a in
                                         self.atoms.get_chemical_symbols()]))

        # Compute the ideal gas free energy
        f_ig = free_energy_ideal_gas(vol,
                                     nat,
                                     self.masses,
                                     self.temperature)  # eV/at

        # Compute Uhlenbeck-Ford excess free energy
        f_uf = free_energy_uhlenbeck_ford(nat_tot/vol,
                                          self.p,
                                          self.sigma,
                                          self.temperature)  # eV/at

        # Compute the work between Uhlenbeck-Ford potential and the MLIP
        u_f, lambda_f = np.loadtxt("forward.dat", unpack=True)
        u_b, lambda_b = np.loadtxt("backward.dat", unpack=True)
        int_f = np.trapz(u_f, lambda_f)
        int_b = np.trapz(u_b, lambda_b)
        work = (int_f - int_b) / 2.0  # eV/at

        # Add everything together
        free_energy = f_ig + f_uf + work
        free_energy_corrected = free_energy
        if self.fcorr1 is not None:
            free_energy_corrected += self.fcorr1
        if self.fcorr2 is not None:
            free_energy_corrected += self.fcorr2

        if self.pressure is not None:
            pv = self.pressure / (160.21766208) * vol / nat_tot
        else:
            pv = 0.0

        # write the results
        with open("free_energy.dat", "w") as f:
            header = "#   T [K]     Fe tot [eV/at]     " + \
                      "Fe harm [eV/at]      Work [eV/at]      PV [eV/at]"
            results = f"{self.temperature:10.3f}     " + \
                      f"{free_energy:10.6f}         " + \
                      f"{f_uf:10.6f}          " + \
                      f"{work:10.6f}          " + \
                      f"{pv:10.6}"
            if self.fcorr1 is not None:
                header += "    Delta F1 [eV/at]"
                results += "         {0:10.6f}".format(self.fcorr1)
            if self.fcorr2 is not None:
                header += "    Delta F2 [eV/at]"
                results += "          {0:10.6f}".format(self.fcorr2)
            if self.fcorr1 is not None or self.fcorr2 is not None:
                header += "    Fe corrected [eV/at]"
                results += "           {0:10.6f}".format(free_energy_corrected)
            header += "\n"
            f.write(header)
            f.write(results)

        msg = "Summary of results for this state\n"
        msg += '===========================================================\n'
        msg += "Frenkel-Ladd path integration, " \
               "with an Uhlenbeck-Ford potential reference\n"
        msg += "Reference potential parameters :\n"
        msg += f"      sigma                     {self.sigma}\n"
        msg += f"      p                         {self.p}\n"
        msg += f"Temperature :                   {self.temperature:10.3f} K\n"
        msg += "Volume :                        " + \
               f"{vol/nat_tot:10.3f} angs^3/at\n"
        msg += f"Free energy :                   {free_energy:10.6f} eV/at\n"
        msg += f"Excess work :                   {work:10.6f} eV/at\n"
        msg += f"Ideal gas free energy :         {f_ig:10.6f} eV/at\n"
        msg += f"UF excess free energy :         {f_uf:10.6f} eV/at\n"
        if self.fcorr1 is not None:
            msg += "1st order true pot correction : " + \
                   f"{self.fcorr1:10.6f} eV/at\n"
        if self.fcorr2 is not None:
            msg += "2nd order true pot correction : " + \
                   f"{self.fcorr2:10.6f} eV/at\n"
        if self.fcorr1 is not None or self.fcorr2 is not None:
            msg += "Free energy corrected :         " + \
                   f"{free_energy_corrected:10.6f} eV/at\n"

        if self.fcorr1 is not None or self.fcorr2 is not None:
            return msg, free_energy_corrected
        else:
            if self.pressure is None:
                return msg, free_energy
            else:
                return msg, free_energy + pv

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """
        """
        if self.damp is None:
            self.damp = "$(100*dt)"

        temp = self.temperature
        self.info_dynamics["temperature"] = temp
        if self.langevin:
            langevinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Integrators")
        block("timestep", f"timestep {self.dt / 1000}")
        block("momenta", "velocity all create " +
              f"{temp} {langevinseed} dist gaussian")
        # If we are using Langevin, we want to remove the random part
        # of the forces
        if self.langevin:
            block("rmv_langevin", "fix ff all store/force")
            txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
                  f"{langevinseed} zero yes"
            block("langevin", txt)
            block("nve", "fix f2 all nve")
        else:
            block("nvt", f"fix f1 all nvt temp {temp} {temp} {self.damp}")
        block("compute temp without cm", "compute c1 all temp/com")
        block("fix cm", "fix_modify f1 temp c1")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """
        """
        if self.trajfile:
            block = LammpsBlockInput("dump", "Dumping")
            txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} "
            txt += "id type xu yu zu vx vy vz fx fy fz "
            txt += "element"
            block("dump", txt)
            block("dump_modify1", "dump_modify dum1 append yes")
            txt = "dump_modify dum1 element " + " ".join([p for p in el])
            block("dump_modify2", txt)
            return block
        else:
            pass

# ========================================================================== #
    def _get_neti(self):
        """
        """
        blocks = []
        pair_style = self.pair_style.split()
        if len(self.pair_coeff) == 1:
            pair_coeff = self.pair_coeff[0].split()
            hybrid_pair_coeff = " ".join([*pair_coeff[:2],
                                          pair_style[0],
                                          *pair_coeff[2:]])
        else:
            hybrid_pair_coeff = []
            for pc in self.pair_coeff:
                pc_ = pc.split()
                hpc_ = " ".join([*pc_[:2], *pc_[2:]])
                hybrid_pair_coeff.append(hpc_)

        block0 = LammpsBlockInput("ufm params", "UF potential parameters")
        block0("p", f"variable p equal {self.p}")
        block0("temp", f"variable T equal {self.temperature}")
        block0("kB", "variable kB equal 8.6173303e-5")
        block0("eps", "variable eps equal ${T}*${p}*${kB}")
        block0("sigma", f"variable sig equal {self.sigma}")
        block0("rc", "variable rc equal 5.0*${sig}")
        blocks.append(block0)

        block1 = LammpsBlockInput("eq fwd",
                                  "Equilibration without UF potential")
        block1("run eq fwd", f"run {self.nsteps_eq}")
        blocks.append(block1)

        block2 = LammpsBlockInput("fwd", "Forward Integration")
        block2("tau", "variable tau equal ramp(1,0)")
        txt = "variable lambda_true equal " + \
              "v_tau^5*(70*v_tau^4-315*v_tau^3+540*v_tau^2-420*v_tau+126)"
        block2("lambda_true", txt)
        block2("lambda_ufm", "variable lambda_ufm equal 1-v_lambda_true")
        if len(self.pair_coeff) == 1:
            txt = "pair_style hybrid/scaled " + \
                 f"v_lambda_true {self.pair_style} " + \
                  "v_lambda_ufm ufm ${rc}\n"
            block2("scaling pair_style", txt)
            txt = "pair_coeff " + hybrid_pair_coeff
            block2("true_pair_coeff", txt)
            block2("ufm_pair_coeff", "pair_coeff * * ufm ${eps} ${sig}")
        else:
            # pair_style comd compatible only with one zbl, To be fixed
            txt = "pair_style hybrid/scaled " + \
                  f"{pair_style[1]} {pair_style[2]} " + \
                  f"{pair_style[3]} v_lambda_true " + \
                  f"{pair_style[4]} v_lambda_ufm ufm ${{rc}}\n"
            block2("scaling pair_style", txt)
            txt = "pair_coeff" + hybrid_pair_coeff[0]
            block2("true_pair_coeff_1", txt)
            txt = "pair_coeff" + hybrid_pair_coeff[1]
            block2("true_pair_coeff_1", txt)
            block2("ufm_pair_coeff", "pair_coeff * * ufm ${eps} ${sig}")

        if len(self.pair_coeff) == 1:
            block2("compute pair true", f"compute c2 all pair {pair_style[0]}")
            block2("compute pair ufm", "compute c3 all pair ufm")
            block2("dU", "variable dU equal (c_c2-c_c3)/atoms")
        else:
            # pair_style comd compatible only with one zbl, To be fixed
            block2("compute pair 1", f"compute c2 all pair {pair_style[1]}")
            block2("compute pair 2", f"compute c4 all pair {pair_style[4]}")
            block2("compute pair ufm", "compute c3 all pair ufm")
            block2("dU", "variable dU equal ((c_c2+c_c4)-c_c3)/atoms")

        block2("lamb", "variable lamb equal 1-v_lambda_true")
        block2("write fwd", "fix  f3 all print 1 \"${dU}  ${lamb}\" title " +
                            "\"# dU lambda\" screen no append forward.dat")
        block2("run", f"run {self.nsteps}")
        block2("unfix write fwd", "unfix f3")
        blocks.append(block2)

        block3 = LammpsBlockInput("eq bwd",
                                  "Equilibration with only UF potential")
        block3("run eq bwd", f"run {self.nsteps_eq}")
        blocks.append(block3)

        block4 = LammpsBlockInput("bwd", "Backrward Integration")
        block4("tau", "variable tau equal ramp(0,1)")
        txt = "variable lambda_true equal " + \
              "v_tau^5*(70*v_tau^4-315*v_tau^3+540*v_tau^2-420*v_tau+126)"
        block4("lambda_true", txt)
        block4("lambda_ufm", "variable lambda_ufm equal 1-v_lambda_true")
        if len(self.pair_coeff) == 1:
            txt = "pair_style hybrid/scaled " + \
                  f"v_lambda_true {self.pair_style} " + \
                  "v_lambda_ufm ufm ${rc}\n"
            block4("scaling pair_style", txt)
            txt = "pair_coeff " + hybrid_pair_coeff
            block4("true_pair_coeff", txt)
            block4("ufm_pair_coeff", "pair_coeff * * ufm ${eps} ${sig}")
        else:
            # pair_style comd compatible only with one zbl, To be fixed
            txt = "pair_style hybrid/scaled " + \
                  f"{pair_style[1]} {pair_style[2]} " + \
                  f"{pair_style[3]} v_lambda_true " + \
                  f"{pair_style[4]} v_lambda_ufm ufm ${{rc}}\n"
            block4("scaling pair_style", txt)
            txt = "pair_coeff " + hybrid_pair_coeff[0]
            block4("true_pair_coeff_1", txt)
            txt = "pair_coeff " + hybrid_pair_coeff[1]
            block4("true_pair_coeff_1", txt)
            block4("ufm_pair_coeff", "pair_coeff * * ufm ${eps} ${sig}")
        block4("write bwd", "fix  f3 all print 1 \"${dU}  ${lamb}\" title " +
                            "\"# dU lambda\" screen no append backward.dat")
        blocks.append(block4)

        return blocks

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        if self.damp is None:
            damp = 100 * self.dt

        msg = "Thermodynamic Integration using " + \
              "Frenkel-Ladd path and Uhlenberg-Ford " + \
              "potential for the liquid state\n"
        msg += f"Temperature :                   {self.temperature}\n"
        msg += f"Langevin damping :              {damp} fs\n"
        msg += f"Timestep :                      {self.dt} fs\n"
        msg += f"Number of steps :               {self.nsteps}\n"
        msg += f"Number of equilibration steps : {self.nsteps_eq}\n"
        msg += "Parameters for UF potential :\n"
        msg += f"      sigma                     {self.sigma}\n"
        msg += f"      p                         {self.p}\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ase.calculators.calculator import Calculator


system_changes = ['positions', 'numbers', 'cell', 'pbc']


# ========================================================================== #
# ========================================================================== #
class MlipCalculator(Calculator):
    """
    Ase Caculator object for MLACS generated MLIP that can take atomic
    configurations from an Atoms object and calculate the energy, forces and
    also stresses.
    """
    implemented_properties = ['energy', 'forces', 'stress']

    def __init__(self, model, **kwargs):
        """
        """
        Calculator.__init__(self, **kwargs)
        self.model = model

# ========================================================================== #
    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        energy, forces, stress = self.model.predict(atoms)

        self.results['energy'] = energy
        self.results['forces'] = forces.reshape(len(atoms), 3)
        self.results['stress'] = stress



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from ase.calculators.lammpsrun import LAMMPS
from ase.units import GPa

from .state import StateManager
from ..core.manager import Manager

default_parameters = {}


# ========================================================================== #
# ========================================================================== #
class OptimizeAseState(StateManager):
    """
    Class to manage Structure optimization with ASE Optimizers.

    Parameters
    ----------
    optimizer: :class:`ase.optimize`
        Optimizer from ase.optimize.
        Default :class:`BFGS`

    opt_parameters: :class:`dict`
        Dictionnary with the parameters for the Optimizer.
        Default: {}

    constraints: :class:`ase.constraints` or :class:`list`
        Constraints to apply to the system during the minimization or list
        of constraints.
        By default there is no constraints.

    cstr_parameters: :class:`dict` or :class:`list`
        Dictionnary with the parameter for the constraints or list
        of constraints parameters.
        Default: {}

    filters: :class:`ase.filters`
        Filters are constraints to apply on the cell during the minimization.
        By default there is no filters.

    fltr_parameters: :class:`dict`
        Dictionnary with the parameter for the constraints.
        Default: {}

    fmax: :class:`float`
        The maximum value for the forces to be considered converged.
        Default: 1e-5

    Examples
    --------

    >>> from ase.io import read
    >>> initial = read('A.traj')
    >>>
    >>> from mlacs.state import OptimizeAseState
    >>> opt = OptimizeAseState()
    >>> opt.run_dynamics(initial, mlip.pair_style, mlip.pair_coeff)

    To perform volume optimization, import the UnitCellFilter filter

    >>> from ase.filters import UnitCellFilter
    >>> opt = OptimizeAseState(filters=UnitCellFilter,
                               fltr_parameters=dict(cell_factor=10))
    >>> opt.run_dynamics(initial, mlip.pair_style, mlip.pair_coeff)
    """
    def __init__(self, optimizer=None, opt_parameters={},
                 constraints=None, cstr_parameters={},
                 filters=None, fltr_parameters={},
                 fmax=1e-5, nsteps=1000, nsteps_eq=100, **kwargs):

        super().__init__(nsteps=nsteps, nsteps_eq=nsteps_eq, **kwargs)

        self._opt = optimizer
        self.criterions = fmax
        self._opt_parameters = default_parameters
        self._opt_parameters.update(opt_parameters)
        if optimizer is None:
            from ase.optimize import BFGS
            self._opt = BFGS

        # RB constraints on Atoms.
        self._cstr = constraints
        self._cstr_params = cstr_parameters

        # RB constraints on Cell.
        self._fltr = filters
        self._fltr_params = fltr_parameters

        self.ispimd = False
        self.isrestart = False

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def run_dynamics(self,
                     supercell,
                     pair_style,
                     pair_coeff,
                     model_post,
                     atom_style="atomic",
                     eq=False,
                     elements=None):
        """
        Run state function.
        """
        atoms = supercell.copy()
        calc = LAMMPS(pair_style=pair_style, pair_coeff=pair_coeff,
                      atom_style=atom_style)
        if model_post is not None:
            calc.set(model_post=model_post)
        atoms.calc = calc
        if eq:
            nsteps = self.nsteps_eq
        else:
            nsteps = self.nsteps

        atoms = self.run_optimize(atoms, nsteps)
        return atoms.copy()

# ========================================================================== #
    def run_optimize(self, atoms, steps):
        """
        Run state function.
        """

        opt_at = atoms
        if self._cstr is not None:
            if isinstance(self._cstr, list):
                cstr = [c(**p) for c, p in zip(self._cstr, self._cstr_params)]
                opt_at.set_constraint(cstr)
            else:
                opt_at.set_constraint(self._cstr(**self._cstr_params))
        if self._fltr is not None:
            opt_at = self._fltr(atoms, **self._fltr_params)

        opt = self._opt(opt_at, **self._opt_parameters)
        opt.run(steps=steps, fmax=self.criterions)

        if self._fltr is not None:
            atoms = opt.atoms.atoms
        else:
            atoms = opt.atoms

        return atoms.copy()

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        msg = "Geometry optimization as implemented in ASE\n"
        # RB not implemented yet.
        # AC now it's implemented, but not easily accessible
        if self._fltr is not None:
            if self._fltr.__name__ == "UnitCellFilter":
                if "scalar_pressure" in self._fltr_params.keys():
                    press = self._fltr_params["scalar_pressure"] / GPa
                else:
                    press = 0.0 / GPa
                msg += f"   target pressure: {press} GPa\n"
        # if self.pressure is not None:
        #    msg += f"   target pressure: {self.pressure}\n"
        msg += f"   min_style: {self._opt.__name__}\n"
        msg += f"   forces tolerance: {self.criterions}\n"
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import os
import sys
import shlex
import shutil
import importlib.util
from subprocess import Popen, PIPE
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))  # noqa
import mlacs  # noqa


def has_lammps_nompi():
    """
    Returns True if Lammps doesn't have mpi.
    """
    envvar = "ASE_LAMMPSRUN_COMMAND"
    exe = os.environ.get(envvar)
    from mlacs.utilities.io_lammps import get_lammps_command
    exe = get_lammps_command()
    error = b'ERROR: Processor partitions do not match number of allocated'
    if exe is None:
        exe = "lmp_mpi"
    cmd = f"{exe} -h"
    lmp_info = Popen(shlex.split(cmd), stdout=PIPE).communicate()[0]
    if b'REPLICA' not in lmp_info:
        return True
    cmd = f"mpirun -n 2 {exe} -partition 2x1"

    if shutil.which("mpirun") is None or shutil.which(exe) is None:
        return True

    lmp_info = Popen(shlex.split(cmd), stdout=PIPE).communicate()[0]
    if error in lmp_info:
        return True
    return False


def has_mlp():
    """
    Returns True if there is no mlp executable.
    """
    return shutil.which("mlp") is None

def has_pyace():
    """
    Returns True if there is no pyace module
    """
    try:
        import pyace
        return False
    except ImportError:
        return True

def has_pyace():
    """
    Returns True if there is no pyace module
    """
    try:
        import pyace  # noqa
        return False
    except ImportError:
        return True


def has_netcdf():
    """
    Returns True if there is no netCDF4 package.
    """
    if importlib.util.find_spec('netCDF4') is None:
        return True
    return False



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.calculator import CalculatorError
try:
    from icet import ClusterSpace
    from icet.tools.structure_generation import generate_sqs_from_supercells
    isicet = True
except ImportError:
    isicet = False

from . import CalcManager


# ========================================================================== #
# ========================================================================== #
class DlmCalcManager(CalcManager):
    """
    Class for Disorder Local Moment simulation.

    Disorder Local Moment is a method to simulate an antiferromagnetic
    material by imposing periodically a random spin configuration
    by means of Special Quasirandom Structures.

    Parameters
    ----------
    calc: :class:`ase.calculator`
        A ASE calculator object.

    unitcell: :class:`ase.atoms`
        The Atoms object for the unitcell, for which the symmetry will
        be used to create the magnetic sqs.

    supercell: :class:`ase.atoms`
        The supercell with ideal positions

    magnetic_sites: :class:`list`
        List of integers describing the unitcell sites where
        the magnetic atoms are located

    mu_b: :class:`float`
        The initial spin amplitude, imposed before the calculation,
        in Bohr magneton. Default ``1.0``.

    cutoff: :class:`list` of :class:`float`
        The cutoffs for the SQS generation.
        See icet documentation for more information. Default ``[6.0, 4.0]``.

    n_steps: :class:`int` (optional)
        Number of Monte-Carlo steps for the generation of the magnetic SQS.
        Default ``3000``.
    """
    def __init__(self,
                 calc,
                 unitcell,
                 supercell,
                 magnetic_sites,
                 mu_b=1.0,
                 cutoffs=[6.0, 4.0],
                 n_steps=3000,
                 **kwargs):
        CalcManager.__init__(self, calc, **kwargs)

        if not isicet:
            msg = "You need the icet package installed to use a DLM calculator"
            raise ModuleNotFoundError(msg)

        chemsymb = [["N"]] * len(unitcell)
        for i in magnetic_sites:
            chemsymb[i] = ["H", "B"]  # H -> haut et B -> bas
        self.cutoffs = cutoffs
        self.mu_b = mu_b
        self.supercell = supercell.copy()
        self.cs = ClusterSpace(unitcell, cutoffs, chemsymb)
        self.n_steps = n_steps
        self.target_concentrations = {"H": 0.5, "B": 0.5}

# ========================================================================== #
    def compute_true_potential(self,
                               confs,
                               state=None,
                               step=None):
        """
        Compute the energy of given configurations with an ASE calculator,
        using a random spin configuration from the SQS.
        """
        confs = [at.copy() for at in confs]
        result_confs = []
        for at in confs:
            sqs = generate_sqs_from_supercells(self.cs,
                                               [self.supercell],
                                               self.target_concentrations,
                                               n_steps=self.n_steps)
            magmoms = np.zeros(len(self.supercell))
            for i, c in enumerate(sqs.get_chemical_symbols()):
                if c == "H":
                    magmoms[i] = self.mu_b
                if c == "B":
                    magmoms[i] = -self.mu_b
            at.set_initial_magnetic_moments(magmoms)
            at.calc = self.calc
            try:
                at.get_potential_energy()
                energy = at.get_potential_energy()
                forces = at.get_forces()
                stress = at.get_stress()
                sp_calc = SinglePointCalculator(at,
                                                energy=energy,
                                                forces=forces,
                                                stress=stress)
                at.calc = sp_calc
                result_confs.append(at)
            except CalculatorError:
                result_confs.append(None)
        return result_confs

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        name = self.calc.name

        msg = "True potential parameters:\n"
        msg += "Calculator : {0}\n".format(name)
        if hasattr(self.calc, "todict"):
            dct = self.calc.todict()
            msg += "parameters :\n"
            for key in dct.keys():
                msg += "   " + key + f"  {dct[key]}\n"
        msg += "Disorder Local Moment method for antifferomagnetic spin\n"
        msg += f"Initial absolute magnetic moment : {self.mu_b}\n"
        msg += "Cutoffs : " + " ".join([str(c) for c in self.cutoffs]) + "\n"
        msg += "Number of Monte-Carlo steps for the sqs generation : " + \
               f"{self.n_steps}\n"
        msg += "Cluster Space:\n"
        msg += str(self.cs)
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, PR, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import copy
import importlib
import numpy as np
from operator import attrgetter

from ase.atoms import Atoms
from ase.units import Hartree, Bohr

from ..core.manager import Manager
from ..utilities.io_lammps import LammpsBlockInput
from ..utilities.miscellanous import read_distribution_files as read_df

ti_args = ['atoms',
           'pair_style',
           'pair_coeff',
           'temperature',
           'pressure',
           'nsteps',
           'nsteps_eq']


# ========================================================================== #
# ========================================================================== #
class CalcProperty(Manager):
    """
    Parent Class for on the fly property calculations.

    Parameters
    ----------
    method: :class:`str`
    Type of criterion.
        - max, maximum difference between to consecutive step < criterion
        - ave, average difference between to consecutive step < criterion
    Default ``max``

    criterion: :class:`float`
        Stopping criterion value (eV). Default ``0.001``
        Can be ``None``, in this case there is no criterion.

    frequence : :class:`int`
        Interval of Mlacs step to compute the property. Default ``1``
    """

    def __init__(self,
                 args={},
                 state=None,
                 method='max',
                 criterion=0.001,
                 frequence=1,
                 **kwargs):

        self.freq = frequence
        self.stop = criterion
        self.method = method
        self.kwargs = args
        self.isfirst = True
        self.isgradient = True
        self.useatoms = True
        self.needdir = True
        self.label = 'Observable_Label'
        self.shape = None
        self.nc_name = None
        self.nc_dim = None
        self.nc_unit = ''
        if state is not None:
            self.state = copy.deepcopy(state)

# ========================================================================== #
    def _exec(self):
        """
        Dummy execution function.
        """
        raise RuntimeError("Execution not implemented.")

# ========================================================================== #
    @property
    def isconverged(self):
        """
        Check if the property is converged.
        """
        if not isinstance(self.new, np.ndarray):
            self.new = np.r_[self.new]
        if self.isfirst:
            self.old = np.zeros(self.new.shape)
            self.isfirst = False
        check = self._check
        self.old = self.new
        return check

# ========================================================================== #
    @property
    def _check(self):
        """
        Check criterions.
        """
        self.maxf = np.max(np.abs(self.new - self.old))
        if not self.isgradient:
            self.maxf = np.max(np.abs(self.new))
        self.avef = np.average(np.abs(self.new - self.old))
        if self.stop is None:
            return False
        elif self.method == 'max' and self.maxf < self.stop:
            return True
        elif self.method == 'ave' and self.avef < self.stop:
            return True
        else:
            return False

# ========================================================================== #
    def get_atoms(self, atoms):
        """
        If reference configuration needed.
        """
        if isinstance(atoms, Atoms):
            self.atoms = [atoms.copy()]
        else:
            self.atoms = atoms.copy()

# ========================================================================== #
    def __repr__(self):
        """
        Dummy function for the real logger.
        """
        return ""


# ========================================================================== #
# ========================================================================== #
class CalcPafi(CalcProperty):
    """
    Class to set a minimum free energy calculation.
    See :func:`PafiLammpsState.run_dynamics` parameters.
    """

    def __init__(self,
                 args,
                 state=None,
                 method='max',
                 criterion=0.001,
                 frequence=1,
                 **kwargs):
        CalcProperty.__init__(self, args, state, method, criterion, frequence,
                              **kwargs)
        self.state.folder = 'PafiPath_Calculation'

# ========================================================================== #
    @Manager.exec_from_workdir
    def _exec(self):
        """
        Exec a MFEP calculation with lammps. Use replicas.
        """
        self.state.workdir = self.folder
        self.state.subfolder = self.subfolder
        atoms = self.state.path.atoms[0]
        mlip = self.kwargs['mlip']
        self.new = self.state.run_pafipath_dynamics(
                atoms, mlip.pair_style, mlip.pair_coeff)[1]
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = 'Computing the minimum free energy path:\n'
        msg += self.state.log_recap_state()
        msg += 'Free energy difference along the path with previous step:\n'
        msg += f'        - Maximum  : {self.maxf}\n'
        msg += f'        - Averaged : {self.avef}\n\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcNeb(CalcProperty):
    """
    Class to set a NEB calculation.
    See :func:`NebLammpsState.run_dynamics` parameters.
    """

    def __init__(self,
                 args,
                 state=None,
                 method='max',
                 criterion=0.001,
                 frequence=1):
        CalcProperty.__init__(self, args, state, method, criterion, frequence)
        self.state.folder = 'Neb_Calculation'

# ========================================================================== #
    @Manager.exec_from_workdir
    def _exec(self):
        """
        Exec a NEB calculation with lammps. Use replicas.
        """
        self.state.workdir = self.folder
        self.state.subfolder = self.subfolder
        atoms = self.state.atoms[0]
        mlip = self.kwargs['mlip']
        self.state.run_dynamics(atoms, mlip.pair_style, mlip.pair_coeff)
        self.state.extract_NEB_configurations()
        self.new = self.state.spline_energies
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = 'Computing the minimum energy path from a NEB calculation:\n'
        msg += self.state.log_recap_state()
        msg += 'Energy difference along the reaction '
        msg += 'path with previous step:\n'
        msg += f'        - Maximum  : {self.maxf}\n'
        msg += f'        - Averaged : {self.avef}\n\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcRdf(CalcProperty):
    """
    Class to set a radial distribution function calculation.
    """

    def __init__(self,
                 args,
                 state=None,
                 method='max',
                 criterion=0.05,
                 frequence=2):
        CalcProperty.__init__(self, args, state, method, criterion, frequence)

        self.useatoms = True
        self.step = self.state.nsteps_eq
        if 'nsteps' in self.kwargs.keys():
            self.step = self.kwargs['nsteps'] / 10
            self.state.nsteps = self.kwargs['nsteps']
            self.kwargs.pop('nsteps')
        self.filename = 'spce-rdf.dat'
        if 'filename' in self.kwargs.keys():
            self.filename = self.kwargs['filename']
            self.kwargs.pop('filename')
        self.state.folder = 'Rdf_Calculation'

# ========================================================================== #
    @Manager.exec_from_workdir
    def _exec(self):
        """
        Exec a Rdf calculation with lammps.
        """

        from ..utilities.io_lammps import get_block_rdf

        self.state.workdir = self.folder
        self.state.subfolder = self.subfolder
        if self.state._myblock is None:
            block = LammpsBlockInput("Calc RDF", "Calculation of the RDF")
            block("equilibrationrun", f"run {self.step}")
            block("reset_timestep", "reset_timestep 0")
            block.extend(get_block_rdf(self.step, self.filename))
            self.state._myblock = block
        mlip = self.kwargs['mlip']
        self.state.run_dynamics(self.atoms[-1], mlip.pair_style,
                                mlip.pair_coeff)
        self.new = read_df(self.state.subsubdir / self.filename)[0]
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = 'For the radial distribution function g(r):\n'
        msg += self.state.log_recap_state()
        msg += f'        - Maximum  : {self.maxf}\n'
        msg += f'        - Averaged : {self.avef}\n\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcAdf(CalcProperty):
    """
    Class to set the angle distribution function calculation.
    """

    def __init__(self,
                 args,
                 state=None,
                 method='max',
                 criterion=0.05,
                 frequence=5):
        CalcProperty.__init__(self, args, state, method, criterion, frequence)

        self.useatoms = True
        self.step = self.state.nsteps_eq
        if 'nsteps' in self.kwargs.keys():
            self.step = self.kwargs['nsteps'] / 10
            self.state.nsteps = self.kwargs['nsteps']
            self.kwargs.pop('nsteps')
        self.filename = 'spce-adf.dat'
        if 'filename' in self.kwargs.keys():
            self.filename = self.kwargs['filename']
            self.kwargs.pop('filename')
        self.state.folder = 'Adf_Calculation'

# ========================================================================== #
    @Manager.exec_from_workdir
    def _exec(self):
        """
        Exec an Adf calculation with lammps.
        """

        from ..utilities.io_lammps import get_block_adf

        self.state.workdir = self.folder
        self.state.subfolder = self.subfolder
        if self.state._myblock is None:
            block = LammpsBlockInput("Calc ADF", "Calculation of the ADF")
            block("equilibrationrun", f"run {self.step}")
            block("reset_timestep", "reset_timestep 0")
            block.extend(get_block_adf(self.step, self.filename))
            self.state._myblock = block
        mlip = self.kwargs['mlip']
        self.state.run_dynamics(self.atoms[-1], mlip.pair_style,
                                mlip.pair_coeff)
        self.new = read_df(self.state.subsubdir / self.filename)[0]
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = 'For the angle distribution function g(theta):\n'
        msg += self.state.log_recap_state()
        msg += f'        - Maximum  : {self.maxf}\n'
        msg += f'        - Averaged : {self.avef}\n\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcTi(CalcProperty):
    """
    Class to set a nonequilibrium thermodynamic integration calculation.
    See the :class:`ThermodynamicIntegration` classe.

    Parameters
    ----------
    phase: :class:`str`
        Structure of the system: solild or liquid.
        Set either the Einstein crystal as a reference system or the UF liquid.
    """

    def __init__(self,
                 args,
                 phase,
                 state=None,
                 ninstance=None,
                 method='max',
                 criterion=0.001,
                 frequence=10):
        CalcProperty.__init__(self, args, state, method, criterion, frequence)

        self.ninstance = ninstance
        self.phase = phase
        if self.phase == 'solid':
            from mlacs.ti import EinsteinSolidState
        elif self.phase == 'liquid':
            from mlacs.ti import UFLiquidState
#        else:
#            print('abort_unkown_phase')
#            exit(1)
        self.ti_state = {}
        self.kwargs = {}
        for keys, values in args.items():
            if keys in ti_args:
                self.ti_state[keys] = values
            else:
                self.kwargs[keys] = values
        if self.phase == 'solid':
            self.state = EinsteinSolidState(**self.ti_state)
        elif self.phase == 'liquid':
            self.state = UFLiquidState(**self.ti_state)

# ========================================================================== #
    @Manager.exec_from_workdir
    def _exec(self):
        """
        Exec a NETI calculation with lammps.
        """
        from mlacs.ti import ThermodynamicIntegration

        # Creation of ti object ---------------------------------------------
        self.ti = ThermodynamicIntegration(self.state,
                                           self.ninstance,
                                           logfile="TiCheckFe.log")
        self.ti.workdir = self.folder
        self.ti.folder = 'Neti_Calculation'
        self.ti.subfolder = self.subfolder

        # Run the simu ------------------------------------------------------
        self.ti.run()
        # Get Fe ------------------------------------------------------------
        if self.ninstance == 1:
            _, self.new = self.state.postprocess(self.ti.get_fedir())
        elif self.ninstance > 1:
            tmp = []
            for i in range(self.ninstance):
                _, tmp_new = self.state.postprocess(self.ti.get_fedir()
                                                    + f"for_back_{i+1}/")
                tmp.append(tmp_new)
            self.new = np.r_[np.mean(tmp)]
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = 'For the free energy convergence check:\n'
        msg += 'Free energy at this step is: '
        for _ in self.new:
            msg += f' {_:10.6f}'
        msg += '\n'
        msg += f'        - Maximum  : {self.maxf}\n'
        msg += f'        - Averaged : {self.avef}\n\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcExecFunction(CalcProperty):
    """
    Class to execute on the fly a python function and converge on the result.

    Parameters
    ----------
    function: :class:`str` or `function`
        Function to call. If the function is a `str`, you to define the
        module to load the function.

    args: :class:`dict`
        Arguments of the function.

    module: :class:`str`
        Module to load the function.

    useatoms: :class:`bool`
        True if the function is called from an ase.Atoms object.

    nc_unit: :class:`str` (optional)
        Unit of the observable saved in *HIST.nc file.
        These units are derived from the atomic unit system: Bohr, Ha, etc.
        Cf. mlacs.utilities.io_abinit.MlacsHist._set_unit_conventions().
        Default ''.

    ase_unit: :class:`str` (optional)
        Unit of the observable (ASE convention).
        Cf. mlacs.utilities.io_abinit.MlacsHist._set_unit_conventions().
        These units are expected to be the `metal` units, cf. ase.units.py.
        Default ''.
    """

    def __init__(self,
                 function,
                 args={},
                 module=None,
                 use_atoms=True,
                 gradient=False,
                 criterion=0.001,
                 frequence=1,
                 nc_unit='',
                 ase_unit=''):
        CalcProperty.__init__(self, args, None, 'max', criterion, frequence)

        self._func = function
        if module is not None:
            importlib.import_module(module)
            self._function = getattr(module, function)
        self.isfirst = True
        self.needdir = False
        self.use_atoms = use_atoms
        self.isgradient = gradient
        self.label = function
        self.shape = None
        self.nc_unit = nc_unit
        self.ase_unit = ase_unit

# ========================================================================== #
    def _unit_converter(self):
        """
        Convert units from ASE unit convention to Abinit's, i.e.,
        the result self.new of the _exec() routine is expressed in atomic unit
        """
        # TODO: Move this routine to utilities.units.py
        eV2Ha = 1/Hartree
        Ang2Bohr = 1/Bohr
        # Dictionary that maps ASE units to the corresponding multiplication
        # factors that convert them to Abinit units system.
        # Example: if a = 100 eV then a*unit_convert_dict['eV'] is 3.67 Hartree
        unit_convert_dict = {'eV': eV2Ha,
                             'Ang': Ang2Bohr,
                             'eV/Ang': eV2Ha/Ang2Bohr,
                             'Ang^3': Ang2Bohr**3,
                             'eV/Ang^3': eV2Ha/Ang2Bohr**3,
                             }

        # Proceed with the conversion itself
        if hasattr(self, 'new') and self.ase_unit in unit_convert_dict:
            self.new *= unit_convert_dict[self.ase_unit]

# ========================================================================== #
    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            self._function = [getattr(_, self._func) for _ in self.atoms]
            self.new = np.r_[[_f(**self.kwargs) for _f in self._function]]
        else:
            self.new = self._function(**self.kwargs)
        self._unit_converter()
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        property.
        """
        msg = f'Converging on the result of {self._func} function\n'
        if self.isgradient:
            msg += 'Computed with the previous step:\n'
        msg += f'        - Maximum  : {self.maxf}\n'
        return msg


# ========================================================================== #
# ========================================================================== #
class CalcRoutineFunction(CalcExecFunction):
    """
    Class to routinely compute basic thermodynamic observables.

    Parameters
    ----------

    function: :class:`str`
        Name of Lammps function, e.g. `get_kinetic_energy'.

    label: :class:`str`
        Label of the function to be executed, e.g. `Kinetic_Energy`.
        Cf. mlacs.utilities.io_abinit.MlacsHist.nc_routine_conv().

    nc_name: :class:`str` (optional)
        Name of the observable in *HIST.nc file, e.g. `ekin`.
        Cf. mlacs.utilities.io_abinit.MlacsHist.nc_routine_conv().
        This name should follow Abinit conventions as much as possible.
        Default ``None``.

    nc_dim: :class:`str` (optional)
        Name of the dimension of the observable in *HIST.nc file.
        Cf. mlacs.utilities.io_abinit.MlacsHist._set_name_conventions().
        Default ``None``.

    nc_unit: :class:`str` (optional)
        Unit of the observable saved in *HIST.nc file.
        These units are derived from the atomic unit system: Bohr, Ha, etc.
        Cf. mlacs.utilities.io_abinit.MlacsHist._set_unit_conventions().
        Default ''.

    ase_unit: :class:`str` (optional)
        Unit of the observable (ASE convention).
        Cf. mlacs.utilities.io_abinit.MlacsHist._set_unit_conventions().
        These units are expected to be the `metal` units, cf. ase.units.py.
        Default ''.

    weight: :class:`WeightingPolicy` (optional)
        WeightingPolicy class, Default: `None`.
    """

    def __init__(self,
                 function,
                 label,
                 nc_name=None,
                 nc_dim=None,
                 nc_unit='',
                 ase_unit='',
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):
        CalcExecFunction.__init__(self, function, dict(), None, True, gradient,
                                  criterion, frequence, nc_unit, ase_unit)
        self.weight = weight
        self.label = label
        self.needdir = False
        self.nc_name = nc_name
        self.nc_dim = nc_dim

# ========================================================================== #
    def __repr__(self):
        """
        Return a string for the log with informations of the calculated
        routine property.
        """
        name_observable = self.label.lower().replace("_", " ")
        unit = self.nc_unit
        msg = f'Routine computation of the {name_observable}\n'
        if len(self.shape) == 0:
            if len(self.new > 0):
                for idx_state, val in enumerate(self.new):
                    msg += f'        - Value for state {idx_state+1} : '
                    msg += '{:.5e}'.format(val) + ' ' + unit + ' \n'
            else:
                msg += f'        - Value for state 1  : {self.new}\n'
        else:
            # Too big to print, cf. *_HIST.nc file
            msg = ''

        return msg


# ========================================================================== #
# ========================================================================== #
class CalcPressure(CalcRoutineFunction):
    """
    Class to compute the hydrostatic pressure.

    Parameters
    ----------

    weight: :class:`WeightingPolicy`
        WeightingPolicy class, Default: `None`.
    """

    def __init__(self,
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):

        label = 'Pressure'
        nc_name = 'press'
        nc_dim = ('time',)
        nc_unit = 'Ha/Bohr^3'
        ase_unit = 'eV/Ang^3'
        CalcRoutineFunction.__init__(self,
                                     'get_stress',
                                     label,
                                     nc_name,
                                     nc_dim,
                                     nc_unit,
                                     ase_unit)

    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            self._function = [getattr(_, self._func) for _ in self.atoms]
            self.new = np.r_[[-np.mean(_f(**self.kwargs)[:3])
                              for _f in self._function]]
        else:
            self.new = self._function(**self.kwargs)
        self._unit_converter()
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged


# ========================================================================== #
# ========================================================================== #
class CalcAcell(CalcRoutineFunction):
    """
    Class to compute the cell lengths.

    Parameters
    ----------
    weight: :class:`WeightingPolicy`
        WeightingPolicy class, Default: `None`.
    """

    def __init__(self,
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):
        label = 'Acell'
        nc_name = 'acell'
        nc_dim = ('time', 'xyz')
        nc_unit = 'Bohr'
        ase_unit = 'Ang'
        CalcRoutineFunction.__init__(self,
                                     'get_cell_lengths_and_angles',
                                     label,
                                     nc_name,
                                     nc_dim,
                                     nc_unit,
                                     ase_unit)

    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            attr = 'cell.cellpar'
            self._function = [attrgetter(attr)(_) for _ in self.atoms]
            self.new = np.r_[[_f(**self.kwargs)[:3] for _f in self._function]]
        else:
            self.new = self._function(**self.kwargs)
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged


# ========================================================================== #
# ========================================================================== #
class CalcAngles(CalcRoutineFunction):
    """
    Class to compute the cell angles.

    Parameters
    ----------
    weight: :class:`WeightingPolicy`
        WeightingPolicy class, Default: `None`.
    """

    def __init__(self,
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):
        label = 'Angles'
        nc_name = 'angl'
        nc_dim = ('time', 'xyz')
        nc_unit = 'deg'
        ase_unit = 'deg'
        CalcRoutineFunction.__init__(self,
                                     'get_cell_lengths_and_angles',
                                     label,
                                     nc_name,
                                     nc_dim,
                                     nc_unit,
                                     ase_unit)

    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            attr = 'cell.cellpar'
            self._function = [attrgetter(attr)(_) for _ in self.atoms]
            self.new = np.r_[[_f(**self.kwargs)[3:] for _f in self._function]]
        else:
            self.new = self._function(**self.kwargs)
        self._unit_converter()
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged


# ========================================================================== #
# ========================================================================== #
class CalcSpinAt(CalcRoutineFunction):
    """
    Class to obtain the electronic spin-magnetization (as computed by Abinit)
    from ASE's Atoms object.
    See also AbinitNC class in mlacs.utilities.io_abinit.py
    """

    def __init__(self,
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):

        label = 'Electronic_Spin_Magnetization'
        nc_name = 'spinat'
        nc_dim = ('time', 'natom', 'xyz',)
        nc_unit = 'hbar/2'
        ase_unit = 'hbar/2'
        CalcRoutineFunction.__init__(self,
                                     '',
                                     label,
                                     nc_name,
                                     nc_dim,
                                     nc_unit,
                                     ase_unit)

    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            try:
                self.new = np.r_[[_.get_array('spinat') for _ in self.atoms]]
            except KeyError:
                self.new = np.r_[[np.zeros((len(_), 3)) for _ in self.atoms]]
        else:
            self.new = self._function(**self.kwargs)
        self._unit_converter()
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged


# ========================================================================== #
# ========================================================================== #
class CalcElectronicEntropy(CalcRoutineFunction):
    """
    Class to obtain the electronic entropy (as computed by Abinit)
    from ASE's Atoms object.
    """

    def __init__(self,
                 weight=None,
                 gradient=False,
                 criterion=None,
                 frequence=1):

        label = 'Electronic_Entropy'
        nc_name = 'entropy'
        nc_dim = ('time',)
        nc_unit = ''
        ase_unit = ''
        CalcRoutineFunction.__init__(self,
                                     '',
                                     label,
                                     nc_name,
                                     nc_dim,
                                     nc_unit,
                                     ase_unit)

    def _exec(self):
        """
        Execute function
        """
        if self.use_atoms:
            try:
                self.new = np.r_[[_.get_properties('')['free_energy']
                                  for _ in self.atoms]]
            except KeyError:
                self.new = np.r_[[0.0 for _ in self.atoms]]
        else:
            self.new = self._function(**self.kwargs)
        self._unit_converter()
        if self.isfirst:
            self.shape = self.new[0].shape
        return self.isconverged



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from ase.atoms import Atoms


def write_cfg(filename, atoms, chemmap):
    if isinstance(atoms, Atoms):
        atoms = [atoms]

    with open(filename, "w") as fd:
        for at in atoms:
            nat = len(at)
            pos = at.get_positions()
            cell = at.get_cell()
            chemsymb = np.array(at.get_chemical_symbols())
            forces = at.get_forces()
            energy = at.get_potential_energy()
            stress = at.get_stress() * at.get_volume()
            idx = np.arange(1, nat+1)
            types = np.zeros((nat), dtype=int)
            for i, el in enumerate(chemmap):
                elidx = np.nonzero(chemsymb == el)[0]
                types[elidx] = i

            fd.write("BEGIN_CFG\n")
            fd.write("Size\n")
            fd.write(f"{nat}\n")
            fd.write("Supercell\n")
            for a, b, c in cell:
                fd.write(f"{a:25.20f}  {b:25.20f}  {c:25.20f}\n")
            fd.write("AtomData: id type cartes_x cartes_y cartes_z ")
            fd.write("fx fy fz\n")
            for i, xpos, f in zip(idx, pos, forces):
                fd.write(f"{i} {types[i-1]} {xpos[0]:25.20f} ")
                fd.write(f"{xpos[1]:25.20f} {xpos[2]:25.20f} ")
                fd.write(f"{f[0]:25.20f} {f[1]:25.20f} {f[2]:25.20f}\n")
            fd.write("Energy\n")
            fd.write(f"{energy:25.20f}\n")
            fd.write("PlusStress: xx yy zz yz xz xy\n")
            fd.write(f"{-stress[0]:25.20f} {-stress[1]:25.20f} ")
            fd.write(f"{-stress[2]:25.20f} {-stress[3]:25.20f} ")
            fd.write(f"{-stress[4]:25.20f} {-stress[5]:25.20f}\n")
            fd.write("END_CFG\n")
            fd.write("\n")


def read_cfg_data(filename):
    energy = []
    forces = []
    stress = []
    with open(filename, "r") as fd:
        for line in fd:
            if line.startswith("BEGIN_CFG"):
                fd.readline()  # Size
                nat = int(fd.readline())
                fd.readline()  # Supercell
                cell = np.zeros((3, 3))
                for i in range(3):
                    line = fd.readline()
                    cell[i] = [float(a) for a in line.split()]
                # We need a dummy Atoms to get the volume for the stress
                atoms = Atoms(["Si"], np.zeros((1, 3)), cell=cell)
                fd.readline()  # AtomData
                pos = np.empty((nat, 3))
                fconf = np.empty((nat, 3))
                for i in range(nat):
                    line = fd.readline().split()
                    pos[i] = [float(a) for a in line[2:5]]
                    fconf[i] = [float(a) for a in line[5:8]]
                fd.readline()  # Energy
                econf = float(fd.readline()) / nat
                fd.readline()  # PlusStress
                line = fd.readline().split()
                sconf = np.array([-float(a) for a in line])
                sconf /= atoms.get_volume()

                energy.append(econf)
                forces.extend(fconf.flatten())
                stress.extend(sconf)
    energy = np.array(energy)
    forces = np.array(forces)
    stress = np.array(stress)
    return energy, forces, stress



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
from .otf_mlacs import OtfMlacs
from .mlminimizer import MlMinimizer

from .version import __version__

__all__ = ['OtfMlacs', 'MlMinimizer', '__version__']



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
from ase.units import GPa

from .mlas import Mlas
from .properties import PropertyManager, CalcExecFunction


# ========================================================================== #
# ========================================================================== #
class MlMinimizer(Mlas):
    r"""
    Class to perform structure minimization assisted with machine-learning
    potential

    Parameters
    ----------

    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
        the atom object on which the simulation is run.

    state: :class:`StateManager` or :class:`list` of :class:`StateManager`
        Object determining the state to be sampled

    calc: :class:`ase.calculators` or :class:`CalcManager`
        Class controlling the potential energy of the system
        to be approximated.
        If a :class:`ase.calculators` is attached, the :class:`CalcManager`
        is automatically created.

    mlip: :class:`MlipManager` (optional)
        Object managing the MLIP to approximate the real distribution
        Default is a LammpsMlip object with a snap descriptor,
        ``5.0`` angstrom rcut with ``8`` twojmax.

    etol: :class:`float` (optional)
        The tolerance for the energy, in eV/at.
        If the difference of energy of the true potential between two
        consecutive steps is lower than ``etol``, the algorithm stops.

    ftol: :class:`float` (optional)
        The tolerance for the forces, in eV/angstrom.
        If the maximum absolute force of the true potential between two
        consecutive steps is lower than ``etol``, the algorithm stops.

    stol: :class:`float` (optional)
        The tolerance for the stress, in GPa.
        If the difference in the stress tensor of the true potential
        between two consecutive steps is lower than ``etol``,
        the algorithm stops.

    neq: :class:`int` (optional)
        The number of equilibration iteration. Default ``10``.

    confs_init: :class:`int` or :class:`list` of :class:`ase.Atoms` (optional)
        if :class:`int`: Number of configuirations used
        to train a preliminary MLIP
        The configurations are created by rattling the first structure
        if :class:`list` of :class:`ase.Atoms`: The atoms that are to be
        computed in order to create the initial training configurations
        Default ``None``.

    std_init: :class:`float` (optional)
        Variance (in :math:`\mathring{a}^2`) of the displacement
        when creating initial configurations.
        Default :math:`0.05 \mathring{a}^2`

    keep_tmp_mlip: :class:`Bool` (optional)
        Keep every generated MLIP. If True and using MBAR, a restart will
        recalculate every previous MLIP.weight using the old coefficients.
        Default ``False``.

    workdir: :class:`str` (optional)
        The directory in which to run the calculation.

    prefix: :class:`str` (optional)
        The prefix to prepend the name of the States files.

    ncprefix: :class:`str` (optional)
        The prefix to prepend the name of the *HIST.nc file.
        Script name format: ncprefix + scriptname + '_HIST.nc'.
        Default `''`.

    ncformat: :class:`str` (optional)
        The format of the *HIST.nc file. One of the five flavors of netCDF
        files format available in netCDF4 python package: 'NETCDF3_CLASSIC',
        'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
        'NETCDF4'.
        Default ``NETCDF3_CLASSIC``.
    """

    def __init__(self, atoms, state, calc, mlip=None, etol=1e-2, ftol=1e-4,
                 stol=1e-3, neq=10, confs_init=None, std_init=0.05,
                 keep_tmp_mlip=False, workdir='', prefix='Trajectory',
                 ncprefix='', ncformat='NETCDF3_CLASSIC'):
        Mlas.__init__(self, atoms, state, calc, mlip=mlip, prop=None, neq=neq,
                      confs_init=confs_init, std_init=std_init,
                      keep_tmp_mlip=keep_tmp_mlip, workdir=workdir,
                      prefix=prefix, ncprefix=ncprefix, ncformat=ncformat)

        if set((etol, ftol, stol)) == {None}:
            msg = "You need to set at least one of etol, ftol or stol"
            raise ValueError(msg)

        prop = []
        if etol is None:
            # We set to stupidly high number so that it's still computed
            etol = 1e8
        prop.append(CalcExecFunction("get_potential_energy",
                                     criterion=etol, frequence=1,
                                     gradient=True))
        if ftol is None:
            # We set to stupidly high number so that it's still computed
            ftol = 1e8
        prop.append(CalcExecFunction("get_forces", criterion=ftol,
                                     frequence=1))
        if stol is None:
            # We set to stupidly high number so that it's still computed
            stol = 1e8
        prop.append(CalcExecFunction("get_stress", criterion=stol * GPa,
                                     frequence=1, gradient=True))

        self.prop = PropertyManager(prop)
        self._etol = etol
        self._ftol = ftol

        msg = "\n\nStructure optimization assisted by machine-learning\n"
        msg += f"   Energy tolerance: {etol} eV/at\n"
        msg += f"   Forces tolerance: {ftol} eV/angs\n"
        msg += f"   Stress tolerance: {stol} GPa\n"
        self.log.logger_log.info(msg)

        self.step = 0
        self.log.logger_log.info("")

# ========================================================================== #
    def _compute_properties(self):
        """
        Main method to compute/save properties of OtfMlacs objects.
        """
        self.prop.calc_initialize(atoms=self.atoms)
        msg = self.prop.run(self.step)
        msg = ""
        ediff = self.prop.manager[0].maxf
        msg += f"Energy difference : {ediff:6.5} eV/at\n"
        maxf = self.prop.manager[1].maxf
        msg += f"Maximum absolute force : {maxf:6.5} eV/angs\n"
        maxs = self.prop.manager[2].maxf / GPa
        msg += f"Maximum stress tensor difference : {maxs:6.5} GPa\n"
        msg == "\n"
        self.log.logger_log.info(msg)

# ========================================================================== #
    def _check_early_stop(self):
        return self.prop.check_criterion



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from mlacs.properties.property_manager import (PropertyManager,
                                               RoutinePropertyManager,
                                               )
from mlacs.properties.calc_property import (CalcPafi,
                                            CalcNeb,
                                            CalcRdf,
                                            CalcAdf,
                                            CalcTi,
                                            CalcProperty,
                                            CalcExecFunction,
                                            CalcRoutineFunction,
                                            CalcPressure,
                                            CalcAcell,
                                            CalcAngles,
                                            CalcElectronicEntropy,
                                            )


__all__ = ['PropertyManager',
           'RoutinePropertyManager',
           'CalcProperty',
           'CalcPafi',
           'CalcNeb',
           'CalcRdf',
           'CalcAdf',
           'CalcTi',
           'CalcExecFunction',
           'CalcRoutineFunction',
           'CalcPressure',
           'CalcAcell',
           'CalcAngles',
           'CalcElectronicEntropy',
           ]



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .state import StateManager
from .langevin import LangevinState
from .lammps_state import LammpsState
from .pafi_lammps_state import PafiLammpsState
from .spin_lammps_state import SpinLammpsState
from .neb_lammps_state import NebLammpsState
from .optimize_lammps_state import OptimizeLammpsState
from .optimize_ase_state import OptimizeAseState
from .mep_ase_state import (LinearInterpolation, NebAseState,
                            CiNebAseState, StringMethodAseState)

__all__ = ['StateManager',
           'LangevinState',
           'LammpsState',
           'PafiLammpsState',
           'SpinLammpsState',
           'NebLammpsState',
           'OptimizeLammpsState',
           'OptimizeAseState',
           'LinearInterpolation',
           'NebAseState',
           'CiNebAseState',
           'StringMethodAseState',
           ]



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from concurrent.futures import ThreadPoolExecutor

import copy
import numpy as np

from ase.units import kB, J, kg, m

from .lammps_state import LammpsState

from ..core.manager import Manager
from ..utilities import save_cwd
from ..utilities import integrate_points as intgpts
from ..utilities.io_lammps import LammpsBlockInput


# ========================================================================== #
# ========================================================================== #
class PafiLammpsState(LammpsState):
    """
    Class to manage constrained MD along a reaction path using the fix Pafi
    with LAMMPS.

    Parameters
    ----------
    temperature: :class:`float`
        Temperature of the simulation, in Kelvin.

    mep: :class:`NebLammpsState`
        Object contain all informations on the MEP (Minimum Energy Path).

    maxjump: :class:`float`
        Maximum atomic jump authorized for the free energy calculations.
        Configurations with an high `maxjump` will be removed.
        Default ``0.4``

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``1.5`` fs.

    damp: :class:`float` or ``None``

    nsteps : :class:`int` (optional)
        Number of MLMD steps for production runs. Default ``1000`` steps.

    nsteps_eq : :class:`int` (optional)
        Number of MLMD steps for equilibration runs. Default ``100`` steps.

    langevin: :class:`Bool`
        If ``True``, a Langevin thermostat is used.
        Else, a Brownian dynamic is used.
        Default ``True``

    fixcm : :class:`Bool` (optional)
        Fix position and momentum center of mass. Default ``True``.

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    rng : RNG object (optional)
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    prt : :class:`Bool` (optional)
        Printing options. Default ``True``

    Examples
    --------

    >>> from ase.io import read
    >>> initial = read('A.traj')
    >>> final = read('B.traj')
    >>>
    >>> from mlacs.state import PafiLammpsState, NebLammpsState
    >>> neb = NebLammpsState([initial, final])
    >>> state = PafiLammpsState(temperature=300, mep=neb)
    >>> state.run_dynamics(atoms, mlip.pair_style, mlip.pair_coeff)
    """

    def __init__(self, temperature, mep=None, maxjump=0.4,
                 dt=1.5, damp=None, prt=False, langevin=True,
                 nsteps=1000, nsteps_eq=100, logfile=None, trajfile=None,
                 loginterval=50, blocks=None, **kwargs):

        super().__init__(temperature=temperature, dt=dt, damp=damp,
                         langevin=langevin,
                         nsteps=nsteps, nsteps_eq=nsteps_eq, logfile=logfile,
                         trajfile=trajfile, loginterval=loginterval,
                         blocks=blocks, **kwargs)

        self.temperature = temperature
        self.mep = mep
        if mep is None:
            raise TypeError('A reaction path must be given!')
        self.mep.print = prt

        self.print = prt
        self.maxjump = maxjump

        self._replica = None

# ========================================================================== #
    @Manager.exec_from_path
    def run_dynamics(self,
                     supercell,
                     pair_style,
                     pair_coeff,
                     model_post=None,
                     atom_style="atomic",
                     eq=False,
                     elements=None):
        """
        Run state function.
        """

        # Run NEB calculation.
        self.mep.workdir = self.workdir
        self.mep.folder = self.folder
        self.mep.subfolder = 'TransPath'
        self.mep.run_dynamics(self.mep.patoms.initial,
                              pair_style, pair_coeff,
                              model_post, atom_style, elements)
        supercell = self.mep.patoms.splined
        self.isrestart = False

        # Run Pafi dynamic at xi.
        atoms = LammpsState.run_dynamics(self, supercell,
                                         pair_style, pair_coeff, model_post,
                                         atom_style, eq, elements)
        return atoms.copy()

# ========================================================================== #
    @Manager.exec_from_path
    def run_pafipath_dynamics(self,
                              supercell,
                              pair_style,
                              pair_coeff,
                              model_post=None,
                              atom_style="atomic",
                              ncpus=1,
                              restart=0,
                              xi=None,
                              nsteps=10000,
                              nthrow=2000,
                              elements=None):
        """
        Run full Pafi path.
        """

        if xi is None:
            xi = np.arange(0, 1.01, 0.01)
        nrep = len(xi)

        afname = self.atomsfname
        lfname = self.lammpsfname

        # Run NEB calculation.
        self.mep.workdir = self.folder
        self.mep.folder = 'TransPath'
        self.mep.run_dynamics(self.mep.patoms.initial,
                              pair_style, pair_coeff,
                              model_post, atom_style, elements)
        self.isrestart = False

        # Run Pafi dynamics.
        with save_cwd(), ThreadPoolExecutor(max_workers=ncpus) as executor:
            for rep in range(restart, nrep):
                worker = copy.deepcopy(self)
                worker._replica = rep
                worker.atomsfname = afname + f'.{rep}'
                worker.lammpsfname = lfname + f'.{rep}'
                worker.subfolder = f'PafiPath_{rep}'
                self.mep.patoms.xi = xi[rep]
                atoms = self.mep.patoms.splined.copy()
                atoms.set_pbc([1, 1, 1])
                executor.submit(LammpsState.run_dynamics,
                                *(worker, atoms, pair_style, pair_coeff,
                                  model_post, atom_style, False, elements))
            executor.shutdown(wait=True)

        # Reset some attributes.
        self._replica = None
        self.atomsfname = afname
        self.lammpsfname = lfname
        return self.log_free_energy(xi, nthrow)

# ========================================================================== #
    @Manager.exec_from_path
    def _write_lammps_atoms(self, atoms, atom_style, elements=None):
        """

        """
        filename = self.path / self.atomsfname

        splat = self.mep.patoms.splined
        splR = self.mep.patoms.splR
        splDR = self.mep.patoms.splDR
        splD2R = self.mep.patoms.splD2R

        from ase.calculators.lammps import Prism, convert
        symbol = splat.get_chemical_symbols()
        species = sorted(set(symbol))
        p = Prism(splat.get_cell())
        xhi, yhi, zhi, xy, xz, yz = convert(p.get_lammps_prism(),
                                            'distance', 'ASE', 'metal')
        instr = f'#{filename} (written by MLACS)\n\n'
        instr += f'{len(symbol)} atoms\n'
        instr += f'{len(species)} atom types\n'
        instr += f'0 {xhi} xlo xhi\n'
        instr += f'0 {yhi} ylo yhi\n'
        instr += f'0 {zhi} zlo zhi\n'
        if p.is_skewed():
            instr += f'{xy} {xz} {yz}  xy xz yz\n'
        instr += '\nAtoms\n\n'
        for i, r in enumerate(splR):
            strformat = '{:>6} ' + '{:>3} ' + ('{:12.8f} ' * 3) + '\n'
            instr += strformat.format(i+1, species.index(symbol[i]) + 1, *r)
        instr += '\nPafiPath\n\n'
        for i, (r, dr, d2r) in enumerate(zip(splR, splDR, splD2R)):
            strformat = '{:>6} ' + ('{:12.8f} ' * 9) + '\n'
            instr += strformat.format(i+1, *r, *dr, *d2r)
        with open(filename, 'w') as w:
            w.write(instr)

# ========================================================================== #
    def _get_block_init(self, atom_style, pbc, el, masses):
        """

        """
        pbc = "{0} {1} {2}".format(*tuple("sp"[int(x)] for x in pbc))

        block = LammpsBlockInput("init", "Initialization")
        block("units", "units metal")
        block("boundary", f"boundary {pbc}")
        block("atom_style", f"atom_style {atom_style}")
        block("atom_modify", "atom_modify  map array sort 0 0.0")
        txt = "neigh_modify every 2 delay 10" + \
              " check yes page 1000000 one 100000"
        block("neigh_modify", txt)
        txt = "fix pat all property/atom d_nx d_ny d_nz" + \
              " d_dnx d_dny d_dnz d_ddnx d_ddny d_ddnz"
        block("property_atoms", txt)
        txt = f"read_data {self.atomsfname} fix pat NULL PafiPath"
        block("read_data", txt)
        for i, mass in enumerate(masses):
            block(f"mass{i}", f"mass {i+1}  {mass}")
        return block

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """

        """
        temp = self.temperature
        seed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("pafi", "Pafi dynamic")

        block("timestep", f"timestep {self.dt / 1000}")
        block("thermo", "thermo 1")
        block("min_style", "min_style fire")  # RB test if we can modify
        txt = "compute cpat all property/atom d_nx d_ny d_nz " + \
              "d_dnx d_dny d_dnz d_ddnx d_ddny d_ddnz"
        block("c_pat", txt)
        block("run_compute", "run 0")

        # RB
        # If we are using Langevin, we want to remove the random part
        # of the forces. RB don't know if i have to do it.
        # if self.langevin:
        #     block("rmv_langevin", "fix ff all store/force")

        if self.langevin:
            txt = f"fix pafi all pafi cpat {temp} {self.damp} {seed} " + \
                  "overdamped no com yes"
            block("langevin", txt)
        else:
            txt = f"fix pafi all pafi cpat {temp} {self.damp} {seed} " + \
                  "overdamped yes com yes"
            block("brownian", txt)
        block("run_fix", "run 0")
        block("minimize", "minimize 0 0 250 250")
        block("reset_timestep", "reset_timestep 0")
        return block

# ========================================================================== #
    def _get_block_custom(self):
        """

        """
        _rep = self._replica
        if self._replica is None:
            _rep = 0

        block = LammpsBlockInput("pafilog", "Pafi log files")
        block("v_dU", "variable dU equal f_pafi[1]")
        block("v_dUe", "variable dUerr equal f_pafi[2]")
        block("v_psi", "variable psi equal f_pafi[3]")
        block("v_err", "variable err equal f_pafi[4]")
        block("c_disp", "compute disp all displace/atom")
        block("c_maxdisp", "compute maxdisp all reduce max c_disp[4]")
        block("v_maxjump", "variable maxjump equal sqrt(c_maxdisp)")
        txt = 'fix pafilog all print 1 ' + \
              '"${dU}  ${dUerr} ${psi} ${err} ${maxjump}" file ' + \
              f'pafi.{_rep}.log title "# dU/dxi (dU/dxi)^2 psi err maxjump"'
        block("pafilog", txt)
        return block

# ========================================================================== #
    @Manager.exec_from_workdir
    def log_free_energy(self, xi, nthrow=2000, _ref=0):
        """
        Extract the MFEP gradient from log files.
        Integrate the MFEP and compute the Free energy barrier.
        """
        temp = self.temperature
        meff = self.mep.patoms.masses

        self.pafi = []
        for rep in range(len(xi)):
            self.subfolder = f'PafiPath_{rep}'
            self.prefix = f'path.{rep}'
            logfile = str(self.get_filepath('.log'))
            data = np.loadtxt(logfile).T[:, nthrow:].tolist()
            self.pafi.append(data)
        self.pafi = np.array(self.pafi)

        dF = []
        psi = []
        cor = []
        maxjump = []
        ntot = len(self.pafi[rep, 0])
        for rep in range(len(xi)):
            # Remove steps with high jumps, the default value is 0.4.
            mj = self.pafi[rep, 4].tolist()
            dF.append(np.average([self.pafi[rep, 0, i]
                      for i, x in enumerate(mj) if x < self.maxjump]))
            psi.append(np.average([self.pafi[rep, 2, i]
                       for i, x in enumerate(mj) if x < self.maxjump]))
            cor.append(np.average([np.log(np.abs(
                       self.pafi[rep, 2, i] / self.pafi[_ref, 2, i]))
                       for i, x in enumerate(mj) if x < self.maxjump]))
            maxjump.append([x for x in mj if x > self.maxjump])
#            dF.append(np.average(self.pafi[rep, 0]))
#            psi.append(np.average(self.pafi[rep, 2]))
#            cor.append(np.average(
#                np.log(np.abs(self.pafi[rep, 2] / self.pafi[_ref, 2]))))
        dF = np.array(dF)
        cor = np.array(cor)
        psi = np.array(psi)
        maxjump = np.array(maxjump)
        F = -np.array(intgpts(xi, dF, xi))
        int_xi = np.linspace(xi[0], xi[F.argmax()], len(xi)//2)
        v = np.array(intgpts(xi, np.exp(- F / kB * temp), int_xi))
        vo = np.sqrt((kB * temp * J) / (2 * np.pi * meff * kg)) / (v[-1] * m)
        Fcor = -np.array(intgpts(xi, dF + kB * temp * cor, xi))
        # Ipsi = np.array(intgpts(xi, psi, xi))
        txt = f'##  Max free energy: {max(F)} eV | frequency: {vo} s-1 | ' + \
              f'effective mass: {meff} uma\n' + \
              '##  xi <dF/dxi> <F(xi)> <psi> cor Fcor(xi) v(xi) NConf ##\n'
        with open(self.workdir / 'free_energy.dat', 'w') as w:
            w.write(txt)
            strformat = ('{:18.10f} ' * 6) + ' {} {}\n'
            for i in range(len(xi)):
                _v = v[-1]
                if i < len(v):
                    _v = v[i]
                w.write(strformat.format(xi[i], dF[i], F[i], psi[i],
                                         kB * temp * cor[i], Fcor[i], _v,
                                         ntot - len(maxjump[i])))
        return np.r_[[F, Fcor, _v]]

# ========================================================================== #
    def log_recap_state(self):
        """
        Function to return a string describing the state for the log
        """
        damp = self.damp
        if damp is None:
            damp = 10 * self.dt
        if self.mep.patoms._xi is None and self.mep.patoms.mode == 'saddle':
            xi = None
        else:
            xi = self.mep.patoms.xi

        msg = self.mep.log_recap_state()
        msg += "Constrained dynamics as implemented in LAMMPS with fix PAFI\n"
        msg += f"Temperature (in Kelvin) :                {self.temperature}\n"
        msg += f"Number of MLMD equilibration steps :     {self.nsteps_eq}\n"
        msg += f"Number of MLMD production steps :        {self.nsteps}\n"
        msg += f"Timestep (in fs) :                       {self.dt}\n"
        msg += f"Themostat damping parameter (in fs) :    {damp}\n"
        if isinstance(xi, float):
            msg += f"Path coordinate :                        {xi}\n"
        elif xi is None:
            msg += "Path coordinate :                        Automatic\n"
        else:
            step = xi[1]-xi[0]
            i, f = (xi[0], xi[-1])
            msg += f"Path interval :                          [{i} : {f}]\n"
            msg += f"Path step interval :                     {step}\n"
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (PR, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.mdMana
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
try:  # Scipy >= 1.6.0
    from scipy.integrate import cumulative_trapezoid as cumtrapz
except ImportError:  # Scipy < 1.6.0
    from scipy.integrate import cumtrapz

from ase.units import kB

from ..core.manager import Manager
from ..utilities.io_lammps import LammpsBlockInput

from .thermostate import ThermoState
from .solids import EinsteinSolidState
from .liquids import UFLiquidState
from .thermoint import ThermodynamicIntegration


# ========================================================================== #
# ========================================================================== #
class ReversibleScalingState(ThermoState):
    """
    Class for performing thermodynamic integration for a
    range of temperature using reversible scaling.

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        ASE atoms object on which the simulation will be performed

    pair_style: :class:`str`
        pair_style for the LAMMPS input

    pair_coeff: :class:`str` or :class:`list` of :class:`str`
        pair_coeff for the LAMMPS input

    fcorr1: :class:`float` or ``None``
        First order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    fcorr2: :class:`float` or ``None``
        Second order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    t_start: :class:`float` (optional)
        Initial temperature of the simulation, in Kelvin. Default ``300``.

    t_end: :class:`float` (optional)
        Final temperature of the simulation, in Kelvin. Default ``1200``.

    fe_init: :class:`float` (optional)
        Free energy of the initial temperature, in eV/at. Default ``None``.

    phase: :class:`str`
        The phase of the system for which the free energy is computed.
        This input is used to compute the reference free energy at the
        starting pressure using a EinsteinSolidState object for solids
        and a UFLiquidState object for liquids. Can be either 'solid'
        or 'liquid'

    ninstance: :class:`int` (optional)
        If Free energy calculation has to be done before temperature sweep,
        settles the number of forward and backward runs. Default ``1``.

    dt: :class:`int` (optional)
        Timestep for the simulations, in fs. Default ``1.5``

    damp : :class:`float` (optional)
        Damping parameter. If ``None``, a damping parameter of a
        hundred time the timestep is used.

    pressure: :class:`float` or ``None``
        Pressure of the simulation.
        If ``None``, simulations are performed in the NVT ensemble.
        Default ``None``.

    pdamp : :class:`float` (optional)
        Damping parameter for the barostat. Default 1000 times ``dt`` is used.
        Default ``None``.

    nsteps: :class:`int` (optional)
        Number of production steps. Default ``10000``.

    nsteps_eq: :class:`int` (optional)
        Number of equilibration steps. Default ``5000``.

    gjf: :class:`bool`
        Whether to use the GJF integrator, if the Langevin thermostat is used.
        Default `True`

    rng: :class:`RNG object`
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    langevin: :class:`bool`
        Whether to use a langevin thermostat. Default `True`

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    interval : :class:`int` (optional)
        Number of steps between log and traj writing. Override
        loginterval. Default ``50``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.
    """
    def __init__(self,
                 atoms,
                 pair_style,
                 pair_coeff,
                 fcorr1=None,
                 fcorr2=None,
                 t_start=300,
                 t_end=1200,
                 fe_init=None,
                 phase=None,
                 ninstance=1,
                 dt=1,
                 damp=None,
                 pressure=None,
                 pdamp=None,
                 nsteps=10000,
                 nsteps_eq=5000,
                 gjf=True,
                 rng=None,
                 langevin=True,
                 logfile=None,
                 trajfile=None,
                 interval=500,
                 loginterval=50,
                 **kwargs):

        self.atoms = atoms
        self.pair_style = pair_style
        self.pair_coeff = pair_coeff
        self.fcorr1 = fcorr1
        self.fcorr2 = fcorr2
        self.t_start = t_start
        self.t_end = t_end
        self.fe_init = fe_init
        self.phase = phase
        self.ninstance = ninstance
        self.damp = damp
        self.pressure = pressure
        self.langevin = langevin
        self.pdamp = pdamp
        self.nsteps = nsteps
        self.nsteps_eq = nsteps_eq
        self.gjf = gjf
        self.dt = dt

        if kwargs.get('folder') is None:
            folder = f"ReversibleScaling_T{self.t_start}K_T{self.t_end}K"
            if self.pressure is None:
                folder += "_NVT"
            else:
                folder += f"_{self.pressure}GPa"
            kwargs['folder'] = folder

        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

        # reversible scaling
        ThermoState.__init__(self,
                             atoms,
                             pair_style,
                             pair_coeff,
                             dt,
                             nsteps,
                             nsteps_eq,
                             rng=rng,
                             logfile=logfile,
                             trajfile=trajfile,
                             interval=interval,
                             loginterval=loginterval,
                             **kwargs)

# ========================================================================== #
    @Manager.exec_from_path
    def run(self):
        """
        """

        if self.fe_init is None and self.phase is not None:
            self.run_single_ti()
        else:
            self.fe_init = 0.0

        self.run_dynamics(self.atoms, self.pair_style, self.pair_coeff)

        with open("MLMD.done", "w") as f:
            f.write("Done")

# ========================================================================== #
    def run_single_ti(self):
        """
        Free energy calculation before sweep
        """
        folder = self.folder + "/Fe_ref"
        if self.phase == 'solid':
            self.state = EinsteinSolidState(self.atoms,
                                            self.pair_style,
                                            self.pair_coeff,
                                            self.t_start,
                                            pressure=self.pressure,
                                            fcorr1=self.fcorr1,
                                            fcorr2=self.fcorr2,
                                            k=None,
                                            dt=self.dt,
                                            workdir=self.workdir,
                                            folder=folder)
        elif self.phase == 'liquid':
            self.state = UFLiquidState(self.atoms,
                                       self.pair_style,
                                       self.pair_coeff,
                                       self.t_start,
                                       pressure=self.pressure,
                                       fcorr1=self.fcorr1,
                                       fcorr2=self.fcorr2,
                                       dt=self.dt,
                                       workdir=self.workdir,
                                       folder=folder)

        self.ti = ThermodynamicIntegration(self.state,
                                           ninstance=self.ninstance,
                                           workdir=self.workdir,
                                           folder=self.folder,
                                           subfolder=self.subfolder,
                                           logfile='FreeEnergy.log')
        self.fe_init = self.ti.run().mean(axis=1)[0]
        return self.fe_init

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """
        """
        if self.damp is None:
            self.damp = "$(100*dt)"

        if self.pdamp is None:
            self.pdamp = "$(1000*dt)"

        temp = self.t_start
        self.info_dynamics["temperature"] = temp
        if self.langevin:
            langevinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Integrators")
        block("timestep", f"timestep {self.dt / 1000}")
        block("momenta", "velocity all create " +
              f"{temp} {langevinseed} dist gaussian")
        # If we are using Langevin, we want to remove the random part
        # of the forces
        if self.langevin:
            block("rmv_langevin", "fix ff all store/force")
            txt = f"fix f1 all langevin {temp} {temp} {self.damp} " + \
                  f"{langevinseed} zero yes"
            block("langevin", txt)
            if self.pressure is None:
                block("nve", "fix f2 all nve")
            if self.pressure is not None:
                # Fix center of mass for barostat
                block("xcm", "variable xcm equal xcm(all,x)")
                block("ycm", "variable ycm equal xcm(all,y)")
                block("zcm", "variable zcm equal xcm(all,z)")
                block("nph", f"fix f2 all nph iso {self.pressure*10000} " +
                             f"{self.pressure*10000} {self.pdamp} " +
                             "fixedpoint ${xcm} ${ycm} ${zcm}")
            block("compute temp without cm", "compute c1 all temp/com")
            block("fix modify cm", "fix_modify f1 temp c1")
            if self.pressure is not None:
                block("fix modify cm", "fix_modify f2 temp c1")
        else:
            if self.pressure is None:
                block("nvt", f"fix f1 all nvt temp {temp} {temp} {self.damp}")
            if self.pressure is not None:
                block("npt", f"fix f1 all npt temp {temp} {temp} " +
                             f"{self.damp} iso {self.pressure*10000} " +
                             f"{self.pressure*10000} {self.pdamp} " +
                             "fixedpoint ${xcm} ${ycm} ${zcm}")
            block("compute temp without cm", "compute c1 all temp/com")
            block("fix modify cm", "fix_modify f1 temp c1")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """
        """
        if self.trajfile:
            block = LammpsBlockInput("dump", "Dumping")
            txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} "
            txt += "id type xu yu zu vx vy vz fx fy fz "
            txt += "element"
            block("dump", txt)
            block("dump_modify1", "dump_modify dum1 append yes")
            txt = "dump_modify dum1 element " + " ".join([p for p in el])
            block("dump_modify2", txt)
            return block
        else:
            pass

# ========================================================================== #
    def _get_neti(self):
        """
        """
        tstart = self.t_start
        tend = self.t_end

        blocks = []
        pair_style = self.pair_style.split()
        if len(self.pair_coeff) == 1:
            pair_coeff = self.pair_coeff[0].split()
            hybrid_pair_coeff = " ".join([*pair_coeff[:2],
                                          pair_style[0],
                                          *pair_coeff[2:]])
        else:
            hybrid_pair_coeff = []
            for pc in self.pair_coeff:
                pc_ = pc.split()
                hpc_ = " ".join([*pc_[:2], *pc_[2:]])
                hybrid_pair_coeff.append(hpc_)

        block0 = LammpsBlockInput("eq fwd", "Equilibration before fwd rs")
        block0("run eq fwd", f"run {self.nsteps_eq}")
        blocks.append(block0)

        block1 = LammpsBlockInput("fwd", "Forward Integration")
        block1("lambda fwd", "variable lambda equal " +
               f"1/(1+(elapsed/{self.nsteps})*({tend}/{tstart}-1))")
        if len(self.pair_coeff) == 1:
            block1("adapt pair_style", "fix f3 all adapt 1 pair " +
                   f"{self.pair_style} scale * * v_lambda")
        else:
            # pair_style comd compatible only with one zbl, To be fixed
            block1("scaling pair_style", "pair_style hybrid/scaled v_lamda " +
                   f"{pair_style[1]} {pair_style[2]} " +
                   f"{pair_style[3]} v_lambda {pair_style[4]}")
            block1("pair_coeff_1", "pair_coeff " + hybrid_pair_coeff[0])
            block1("pair_coeff_2", "pair_coeff " + hybrid_pair_coeff[1])
        block1("write fwd", "fix f4 all print 1 " +
               "\"$(pe/atoms) ${mypress} ${vol} ${lambda}\" screen no " +
               "append forward.dat title " +
               "\"# pe    pressure    vol    lambda\"\n")
        block1("run fwd", f"run {self.nsteps}")
        if len(self.pair_coeff) == 1:
            block1("unfix f3", "unfix f3 ")
        block1("unfix f4", "unfix f4 ")
        blocks.append(block1)

        block2 = LammpsBlockInput("eq bwd", "Equilibration before bwd rs")
        block2("run eq bwd", f"run {self.nsteps_eq}")
        blocks.append(block2)

        block3 = LammpsBlockInput("bwd", "Backward Integration")
        block3("lambda bwd", "variable lambda equal " +
               f"1/(1+(1-elapsed/{self.nsteps})*({tend}/{tstart}-1))")
        if len(self.pair_coeff) == 1:
            block3("adapt pair_style", "fix f3 all adapt 1 pair " +
                   f"{self.pair_style} scale * * v_lambda")
        else:
            # pair_style comd compatible only with one zbl, To be fixed
            block3("scaling pair_style", "pair_style hybrid/scaled v_lamda " +
                   f"{pair_style[1]} {pair_style[2]} " +
                   f"{pair_style[3]} v_lambda {pair_style[4]}")
            block3("pair_coeff_1", "pair_coeff " + hybrid_pair_coeff[0])
            block3("pair_coeff_2", "pair_coeff " + hybrid_pair_coeff[1])
        block3("write bwd", "fix f4 all print 1 " +
               "\"$(pe/atoms) ${mypress} ${vol} ${lambda}\" screen no " +
               "append backward.dat title " +
               "\"# pe    pressure    vol    lambda\"\n")
        blocks.append(block3)
        return blocks

# ========================================================================== #
    @Manager.exec_from_path
    def postprocess(self):
        """
        Compute the free energy from the simulation
        """
        natoms = len(self.atoms)
        if self.pressure is not None:
            p = self.pressure/160.21766208  # already divided by 100000
        else:
            p = 0.0
        # Get data
        v_f, fp, fvol, lambda_f = np.loadtxt("forward.dat", unpack=True)
        v_b, bp, bvol, lambda_b = np.loadtxt("backward.dat", unpack=True)

        v_f /= lambda_f
        v_b /= lambda_b

        # add pressure contribution
        fvol = fvol / natoms
        bvol = bvol / natoms
        v_f = v_f + p * fvol
        v_b = v_b + p * bvol

        # Integrate the forward and backward data
        int_f = cumtrapz(v_f, lambda_f, initial=0)
        int_b = cumtrapz(v_b[::-1], lambda_b[::-1], initial=0)
        # Compute the total work
        work = (int_f + int_b) / (2 * lambda_f)

        temperature = self.t_start / lambda_f
        free_energy = self.fe_init / lambda_f + \
            1.5 * kB * temperature * np.log(lambda_f) + work

        results = np.array([temperature, free_energy]).T
        header = "   T [K]    F [eV/at]"
        fmt = "%10.3f  %10.6f"
        np.savetxt("free_energy.dat", results, header=header, fmt=fmt)
        return

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        npt = False
        if self.pressure is not None:
            npt = True
        if self.damp is None:
            self.damp = 100 * self.dt

        msg = "Thermodynamic Integration using Reversible Scaling\n"
        msg += f"Starting temperature :          {self.t_start}\n"
        msg += f"Stopping temperature :          {self.t_end}\n"
        msg += f"Langevin damping :              {self.damp} fs\n"
        msg += f"Timestep :                      {self.dt} fs\n"
        msg += f"Number of steps :               {self.nsteps}\n"
        msg += f"Number of equilibration steps : {self.nsteps_eq}\n"
        if not npt:
            msg += "Constant volume simulation\n"
        else:
            if self.pdamp is None:
                pdamp = 1000 * self.dt
            msg += "Constant pressure simulation\n"
            msg += f"Pressure :                      {self.pressure} GPa\n"
            msg += f"Pressure damping :              {pdamp} fs\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, PR)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path

import numpy as np
from ase.io import read
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.lammps import Prism, convert

from .path_integral import hbar


class LammpsInput:
    """

    """
    def __init__(self, preambule=None):
        if preambule is not None:
            self.preambule = f"# {preambule}\n\n"
        else:
            self.preambule = ""
        self.nvar = 0
        self.vardict = dict()

    def add_block(self, name, block, order=-1, before=None, after=None):
        """

        """
        if before is not None and after is not None:
            msg = "before and after can't be both set"
            raise ValueError(msg)

        if before is not None:
            order = self.vardict[before]["order"]
        elif after is not None:
            order = self.vardict[after]["order"] + 1

        if order < 0:
            order = self.nvar + 1
        else:
            keys = []
            values = []
            for key, val in self.vardict.items():
                keys.append(key)
                values.append(val["order"])
            keys = np.array(keys)
            values = np.array(values)
            argsort = np.argsort(values)
            values = values[argsort]
            keys = keys[argsort]
            if order > np.max(values) or order not in values:
                self.vardict["name"]["order"] = order
            elif order in values:
                values[values >= order] += 1
                for i, (key, val) in enumerate(zip(keys, values)):
                    self.vardict[key]["order"] = values[i]
        self.vardict[name] = dict(order=order, block=block)
        self.nvar += 1

    def to_string(self):
        """

        """
        keys = []
        orders = []
        blocks = []
        for key, val in self.vardict.items():
            keys.append(key)
            orders.append(val["order"])
            blocks.append(val["block"])

        keys = np.array(keys)
        orders = np.array(orders)
        blocks = np.array(blocks)

        argsort = np.argsort(orders)
        blocks = blocks[argsort]

        txt = self.preambule
        txt += "\n\n".join(str(block) for block in blocks)
        return txt

    def pop(self, name):
        return self.vardict.pop(name)

    def __str__(self):
        return self.to_string()

    def __call__(self, name, block, order=-1):
        self.add_block(name, block, order)


class LammpsBlockInput:
    """

    """
    def __init__(self, name, title=None):
        self.name = name
        self.vardict = dict()
        self.nvar = 0
        if title is not None:
            title = title.strip()
            nchar = len(title)
            self.title = "#" * (12 + nchar) + "\n"
            self.title += title.center(nchar + 10, " ").center(nchar + 12, "#")
            self.title += "\n"
            self.title += "#" * (12 + nchar) + "\n"
        else:
            self.title = "\n"

    def add_variable(self, name, line, order=-1, before=None, after=None):
        """

        """
        if before is not None and after is not None:
            msg = "before and after can't be both set"
            raise ValueError(msg)

        if before is not None:
            order = self.vardict[before]["order"]
        elif after is not None:
            order = self.vardict[after]["order"] + 1

        if order < 0:
            order = self.nvar + 1
        else:
            keys = []
            values = []
            for key, val in self.vardict.items():
                keys.append(key)
                values.append(val["order"])
            keys = np.array(keys)
            values = np.array(values)
            argsort = np.argsort(values)
            values = values[argsort]
            keys = keys[argsort]
            if order in values:
                values[values >= order] += 1
                for i, (key, val) in enumerate(zip(keys, values)):
                    self.vardict[key]["order"] = values[i]
        self.vardict[name] = dict(order=order, line=line)
        self.nvar += 1

    def to_string(self):
        """

        """
        keys = []
        orders = []
        lines = []
        for key, val in self.vardict.items():
            keys.append(key)
            orders.append(val["order"])
            lines.append(val["line"])

        keys = np.array(keys)
        orders = np.array(orders)
        lines = np.array(lines)

        argsort = np.argsort(orders)
        line = lines[argsort]

        txt = self.title
        txt += "\n".join(line)
        return txt

    def pop(self, name):
        '''Remove block line'''
        return self.vardict.pop(name)

    def extend(self, block):
        '''Concatenate Blocks'''
        for key, val in block.vardict.items():
            self.__call__(key, val['line'])

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return f"LammbsBlockInput({self.name})"

    def __call__(self, name, line, order=-1):
        self.add_variable(name, line, order)


class EmptyLammpsBlockInput(LammpsBlockInput):
    """

    """
    def __init__(self, name):
        self.name = name

    def to_string(self):
        return ""


# ========================================================================== #
def get_block_rdf(nsteps, filename='spce-rdf.dat', rmax=None):
    """
    Function to compute and output the radial distribution function
    """
    # freq = int(nsteps/5)
    block = LammpsBlockInput("RDF", "Compute RDF")
    block("v_rep_rdf", f"variable rep_rdf equal {nsteps}/2")
    txt = "compute rdf all rdf ${rep_rdf} 1 1"
    if rmax is not None:
        txt += ' cutoff {rmax}'
    block("c_rdf", txt)
    txt = "fix rdf all ave/time 1 ${rep_rdf}" + \
          f" {nsteps} c_rdf[*] file {filename} mode vector\n"
    block("rdf", txt)
    return block


# ========================================================================== #
def get_block_adf(nsteps, filename='spce-adf.dat'):
    """
    Function to compute and output the angle distribution function
    """
    # freq = int(nsteps/5)
    block = LammpsBlockInput("ADF", "Compute ADF")
    block("v_rep_adf", "variable rep_adf equal 1")
    block("c_adf", "compute adf all adf 360")
    txt = "fix adf all ave/time 100 ${rep_adf}" + \
          f" {nsteps} c_adf[*] file {filename} mode vector\n"
    block("adf", txt)
    return block


# RB: This could be done in a better way.
# ========================================================================== #
def get_block_diffusion(nsteps, filename='diffusion.dat'):
    """
    Function to compute and output the diffusion coefficient
    """
    # freq = int(nsteps/5)
    block = LammpsBlockInput("MSD", "Compute MSD and diffusion coefficient")
    block("v_t", "variable t equal step")
    block("c_msd", "compute msd all msd")
    block("v_msd", "variable msd equal c_msd[4]")
    block("v_twopts", "variable twopoint equal c_msd[4]/6/(step*dt+1.0e-6)")
    block("f_msd", "fix msd all vector 1000 c_msd[4]")
    block("v_slope", "variable fitslope equal slope(f_msd)/6/(10000*dt)")
    txt = 'fix dcoeff all print 100 "${t} ${msd} ${twopoint} ${fitslope}"' + \
          f' append {filename} title "# Step MSD D(start) D(slope)"'
    block("diffusion", txt)
    return block


# ========================================================================== #
def write_atoms_lammps_spin_style(fd, atoms, spin, velocities=True):
    """
    Function to write atoms in the LAMMPS spin style
    Loosely adapted from ASE write_lammpsdata function
    """
    fd.write("# Atoms in spin style, Written by MLACS\n\n")

    nat = len(atoms)
    fd.write(f"{nat} atoms\n")

    symbols = atoms.get_chemical_symbols()
    species = sorted(set(symbols))
    n_atom_type = len(species)
    fd.write(f"{n_atom_type} atom types\n\n")

    prismobj = Prism(atoms.get_cell())
    xhi, yhi, zhi, xy, xz, yz = convert(prismobj.get_lammps_prism(),
                                        'distance',
                                        'ASE',
                                        'metal')

    fd.write(f'0.0 {xhi:23.17g} xlo xhi\n')
    fd.write(f'0.0 {yhi:23.17g} ylo yhi\n')
    fd.write(f'0.0 {zhi:23.17g} zlo zhi\n')
    fd.write("\n\n")

    fd.write("Atoms # spin\n\n")

    pos = prismobj.vector_to_lammps(atoms.get_positions(), wrap=False)
    for i, r in enumerate(pos):
        r = convert(r, "distance", "ASE", "metal")
        s = species.index(symbols[i]) + 1
        line = f"{i+1:>6} {s:>3} "  # Index and species
        line += f"{r[0]:23.17f} {r[1]:23.17f} {r[2]:23.17f} "  # Positions
        norm = np.linalg.norm(spin[i])
        if np.isclose(norm, 0, 1e-5):
            norm = 0.0
            sp = np.zeros(3)
        else:
            sp = spin[i] / norm
        line += f"{sp[0]:23.17f} {sp[1]:23.17f} {sp[2]:23.17f} "
        line += f"{norm} "
        line += "\n"
        fd.write(line)

    if velocities and atoms.get_velocities() is not None:
        fd.write("\n\nVelocities \n\n")
        vel = prismobj.vector_to_lammps(atoms.get_velocities())
        for i, v in enumerate(vel):
            v = convert(v, "velocity", "ASE", "metal")
            fd.write(
                "{0:>6} {1:23.17g} {2:23.17g} {3:23.17g}\n".format(
                    *(i + 1,) + tuple(v)
                )
            )

    fd.flush()


# ========================================================================== #
def reconstruct_mlmd_trajectory(trajfile, logfile):
    """
    Function to reconstruct a trajectory from LAMMPS
    Note that this function is made specifically for
    trajectory launched with MLACS, and won't work
    for general LAMMPS trajectories.

    Parameters
    ----------
    trajfile: Path or str
        The file for the trajectory
    logfile: Path or str
        The file for the log
    return
    ------
        Traj: list of ase.Atoms object
    """
    trajfile = Path(trajfile)
    logfile = Path(logfile)

    isspin = False

    # First we get the index of the quantities we want
    with open(logfile, "r") as fd:
        line = np.array(fd.readline().split())
        assert line[0] == "#"

        idx_epot = np.where(line == "Epot")[0][0] - 1

        if "Magn_x" in line:
            isspin = True
            idx_emagn = np.where(line == "Espin")

    trajconf = read(trajfile, ":")
    log = np.loadtxt(logfile)[:-1]  # the -1 is due to the last dump
    nconf = len(trajconf)
    assert nconf == log.shape[0]

    traj = []
    for at, logat in zip(trajconf, log):
        newat = at.copy()
        epot = logat[idx_epot]
        if isspin:
            espin = logat[idx_emagn]

            sp_n = newat.arrays.pop("c_spin[1]")
            sp_x = newat.arrays.pop("c_spin[2]")
            sp_y = newat.arrays.pop("c_spin[3]")
            sp_z = newat.arrays.pop("c_spin[4]")
            sp = np.c_[sp_x, sp_y, sp_z] * sp_n

            fsp_x = newat.arrays.pop("c_spin[5]")
            fsp_y = newat.arrays.pop("c_spin[6]")
            fsp_z = newat.arrays.pop("c_spin[7]")
            fsp = np.c_[fsp_x, fsp_y, fsp_z] * hbar / (2 * np.pi)

            epot = epot + espin

            newat.set_array("spins", sp)
            newat.set_array("spin_forces", fsp)

        try:
            f_x = newat.arrays.pop("f_ff[1]")
            f_y = newat.arrays.pop("f_ff[2]")
            f_z = newat.arrays.pop("f_ff[3]")
            forces = np.c_[f_x, f_y, f_z]
        except KeyError:
            forces = at.get_forces()

        calc = SinglePointCalculator(newat, energy=epot, forces=forces,
                                     stress=np.zeros(6))
        newat.calc = calc
        traj.append(newat)
    return traj

# RB: msdfile variable probably not needed.
# ========================================================================== #
def get_msd_input(self, msdfile):
    """
    Function to compute msd for neti in solid
    """
    block = LammpsBlockInput("msd", "Compute MSD")
    block("eq", f"run {self.nsteps_eq}")
    for iel, el in enumerate(self.elem):
        block("compute", f"compute c{10+iel} {el} msd com yes")
        block("variable", f"variable msd{el} equal c_c{10+iel}[4]")
        block("msd el", f"fix f{iel+3} {el} print 1 " +
              f"\"${{msd{el}}}\" screen no append msd{el}.dat")
    return block


# ========================================================================== #
def get_lammps_command():
    '''
    Function to load the bash command to run LAMMPS
    '''
    # Since some ASE update, there is a new way to get commands
    envvar = "ASE_LAMMPSRUN_COMMAND"
    try:
        from ase.config import cfg
        if "lammps" in cfg.parser:
            section = cfg.parser["lammps"]
            cmd = section["command"]
        else:
            cmd = cfg.get(envvar)
    except ModuleNotFoundError:
        # The goal is to have this deprecated in the long run
        # when the update of file-io calculators in ASE is completely done
        import os
        cmd = os.environ.get(envvar)

    # And we try the default one afterward
    if cmd is None:
        cmd = "lammps"

    return cmd



# conftest.py

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--full", action="store_true", default=False,
        help="run tests and examples"
    )
    parser.addoption(
        "--fast", action="store_true", default=False,
        help="run tests and examples"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "examples: mark test as examples")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--full"):
        # --full given in cli: do not skip examples and slow tests
        return
    skip = pytest.mark.skip(reason="need --full option to run")
    for item in items:
        if "examples" in item.keywords:
            item.add_marker(skip)
    if not config.getoption("--fast"):
        # --fast given in cli: skip slow tests
        return
    skip = pytest.mark.skip(reason="remove --fast option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs import OtfMlacs
from mlacs.mlip import MliapDescriptor, LinearPotential, DeltaLearningPotential
from mlacs.state import (NebLammpsState,
                         NebAseState,
                         LinearInterpolation,
                         StringMethodAseState)
# from mlacs.properties import CalcNeb


@pytest.fixture
def expected_folder(expected_folder_base):
    folder = expected_folder_base
    folder.append("Mliap")
    folder.pop(folder.index("Snap"))
    return expected_folder_base


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


def issame(a, b):
    return np.all(np.abs(a - b) < 1e-8)


@pytest.mark.slow
@pytest.mark.skipif(context.has_lammps_nompi(),
                    reason="Lammps needs mpi to run PAFI")
def test_mlacs_nebstate_lammps(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 3
    calc = EMT()

    mlip_params = dict(nmax=4, lmax=4)
    desc = MliapDescriptor(atoms, 4.2, mlip_params, style="so3")
    mlip = LinearPotential(desc, folder="Mliap")
    ps, cs = 'zbl 1.0 2.0', ['* * 47 47']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    mode = "rdm_memory"
    nimages = 6
    state = NebLammpsState(nebat, nimages=nimages, mode=mode)

    # RB I will add this later.
    # pair_mlip = dict(pair_style=mlip.pair_style, pair_coeff=mlip.pair_coeff)
    # func = CalcNeb(state=state, args=pair_mlip)
    # sampling = OtfMlacs(nebat[0], state, calc, mlip, func, neq=5)

    # Check that the same Atoms are used
    assert nebat[0] == state.patoms.initial
    assert nebat[-1] == state.patoms.final

    sampling = OtfMlacs(nebat[0], state, calc, dmlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms
    # Check the size of splined objects
    assert nimages == state.patoms.nreplica
    assert len(state.patoms.splE) == 1

    xi = np.linspace(0, 1, 1001)
    state.patoms.xi = xi
    assert len(state.patoms.splE) == len(xi)
    # Check the effective masses for vacancies only one atom should move
    w_m = state.patoms.masses
    assert len(w_m) == natoms
    assert np.sum(w_m) >= 1 and np.sum(w_m) < natoms


def test_mlacs_linear_ase(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 3
    calc = EMT()

    mlip_params = dict(nmax=4, lmax=4)
    desc = MliapDescriptor(atoms, 4.2, mlip_params, style="so3")
    mlip = LinearPotential(desc, folder="Mliap")
    ps, cs = 'zbl 1.0 2.0', ['* * 47 47']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    mode = "rdm_memory"
    nimages = 6
    state = LinearInterpolation(nebat, nimages=6, mode=mode, model=dmlip)

    # Check that the same Atoms are used
    assert nebat[0] == state.patoms.initial
    assert nebat[-1] == state.patoms.final

    sampling = OtfMlacs(nebat[0], state, calc, dmlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms
    # Check the size of splined objects
    assert nimages+2 == state.patoms.nreplica
    assert len(state.patoms.splE) == 1

    xi = np.linspace(0, 1, 1001)
    state.patoms.xi = xi
    assert len(state.patoms.splE) == len(xi)


@pytest.mark.slow
def test_mlacs_neb_ase(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 3
    calc = EMT()

    mlip_params = dict(nmax=4, lmax=4)
    desc = MliapDescriptor(atoms, 4.2, mlip_params, style="so3")
    mlip = LinearPotential(desc, folder="Mliap")
    ps, cs = 'zbl 1.0 2.0', ['* * 47 47']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    nimages = 6
    state = NebAseState(nebat, nimages=6, model=dmlip)

    # Check that the same Atoms are used
    assert nebat[0] == state.patoms.initial
    assert nebat[-1] == state.patoms.final

    sampling = OtfMlacs(nebat[0], state, calc, dmlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms
    # Check the size of splined objects
    assert nimages+2 == state.patoms.nreplica
    assert len(state.patoms.splE) == 1

    xi = np.linspace(0, 1, 1001)
    state.patoms.xi = xi
    assert len(state.patoms.splE) == len(xi)


@pytest.mark.slow
def test_mlacs_string_ase(root, treelink):

    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 3
    calc = EMT()

    mlip_params = dict(nmax=4, lmax=4)
    desc = MliapDescriptor(atoms, 4.2, mlip_params, style="so3")
    mlip = LinearPotential(desc, folder="Mliap")
    ps, cs = 'zbl 1.0 2.0', ['* * 47 47']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    mode = "rdm_memory"
    nimages = 6
    state = StringMethodAseState(nebat, nimages=6, mode=mode, model=dmlip,
                                 ftol=1e-1)

    # Check that the same Atoms are used
    assert nebat[0] == state.patoms.initial
    assert nebat[-1] == state.patoms.final

    sampling = OtfMlacs(nebat[0], state, calc, dmlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    traj = read(root / "Trajectory.traj", ":")
    # Check that the system didn't change in the process
    for at in traj:
        assert len(at) == natoms
    # Check the size of splined objects
    assert nimages+2 == state.patoms.nreplica
    assert len(state.patoms.splE) == 1

    xi = np.linspace(0, 1, 1001)
    state.patoms.xi = xi
    assert len(state.patoms.splE) == len(xi)
